# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

from acris.idioms.data_types import MergedChainedDict
from acris.idioms.decorate import traced_method
from acris.idioms.sequence import Sequence
from acris.idioms.singleton import NamedSingleton
from acris.idioms.threaded import Threaded
import threading
from collections import OrderedDict
import queue
import time
import logging
import inspect
from collections import namedtuple

logger=logging.getLogger(__name__)

traced=traced_method(logger.debug, True)

class ResourcePoolError(Exception): pass
class RequestNotFound(Exception): pass

class Resource(object):  

    def __init__(self, *args, **kwargs):
        '''Instantiate Resource to used in Resource Pool
        
        Args:
            name: resource name
        '''
        self.args=args
        self.kwargs=kwargs
        #self.id_=resource_id_sequence()
        self.resource_name=self.__class__.__name__
        self.pool=None
        
    '''
    def setattrib(self, name, value):
        setattr(self, name, value)
        return self
    '''
        
    def __repr__(self):
        result="Resource(name:%s.%s)" % (self.pool, self.resource_name,)
        return result 
        
Ticket=namedtuple('Ticket', ['pool_name', 'sequence'])


class ResourcePool(NamedSingleton): 
    ''' Singleton pool to managing resources of multiple types.
    
    ResourcePool uses Singleton characteristics to maintain pool per resource name.
    
    Pool policy:
        If there is resource in the pool: provide
        If there is no resource in the pool:
            if limitless, manufacture new by provided resourceFactory
            if limited:
                if nowait: return empty
                if wait with no time limit: wait until resource is available.
                if wait with time limit: halt wait and return with resource if available or empty
                if callback, reserve resource and notify availability with reference to reserved resource
                
    Policy can only be set one immediately after initialization.
    
    '''
    
    __policy = {'autoload': True, # automatically load resources when fall behind
                'load_size': 1, # when autoload, number of resources to load
                'resource_limit': -1, # max resources available
        }
    
    __allow_set_policy=True
    __resource_pool_lock=threading.Lock()
    
    __name=''
    
    __ticket_sequence=Sequence('ResourcePool_ticket')
    __pool_id_sequence=Sequence('ResourcePool_pool_id')
    
    def __init__(self, name='', resource_cls=Resource, policy={}):
        # sets resource pool policy overriding defaults
        self.name=name
        self.__resource_cls=resource_cls
        self.__available_resources=list()
        self.__awaiting=OrderedDict()
        self.__reserved=OrderedDict()
        self.__inuse_resources=list()
        self.__id=self.__pool_id_sequence()
        #self.mutex = threading.RLock()
        #self.__ticket_sequence=Sequence("ResourcePool.%s" % (resource_cls.__name__, ))
        
        if self.__allow_set_policy:
            self.__policy=MergedChainedDict(policy, self.__policy)
        else:
            #self.__lock.release()
            raise ResourcePoolError("ResourcePool already in use, cannot set_policy")
        
    def __repr__(self):
        return "ResourcePool( class: %s, policy: %s)" % (self.__resource_cls.__name__, self.__policy)
    
    @traced
    def __sync_acquire(self):
        #(frame, filename, line_number,
        # function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        #logger.debug("ResourcePool Acquiring Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_lock.acquire()
        
    @traced
    def __sync_release(self):
        #(frame, filename, line_number,
        # function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        #logger.debug("ResourcePool Releasing Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_lock.release()
        
    def __load(self, sync=False, count=-1):
        ''' loads resources into pool
        '''
        if sync: self.__sync_acquire()
        self.__allow_set_policy=False
        if count is None or count < 0:
            load_size=self.__policy['load_size']
            resource_limit=self.__policy['resource_limit']
            count=min(load_size, resource_limit) if resource_limit >=0 else load_size
        
        count=count-len(self.__available_resources) 
        logger.debug('Loading %s resources' % count)
        if count > 0: 
            for _ in range(count):
                resource=self.__resource_cls()
                resource.pool=self.name
                self.__available_resources.append(resource)
        if sync: self.__sync_release()          
    
    def load(self, count=-1):
        ''' loads resources into pool
        '''
        self.__load(sync=True, count=-1)         
        return self
    
    def __remove_ticket(self, ticket, sync=False):
        if sync: self.__sync_acquire()
        logger.debug("Removing reservation ticket: %s, %s" % (ticket, repr(self.__reserved)))
        del self.__reserved[ticket]
        if sync: self.__sync_release()
    
    @Threaded()
    def __wait_on_condition_callback(self, condition, seconds, ticket, callback):
        logger.debug("%s waiting on condition callback: ticket: %s:" % (self.name,ticket,))     
        try:
            with condition:
                condition.wait(seconds)
        except RuntimeError:
            pass
        except Exception as e:
            raise e   
        logger.debug("%s woke up from condition callback: ticket: %s:" % (self.name,ticket,))     
        callback(ticket)
        
    def __wait_on_condition_here(self, condition, seconds, ticket):
        logger.debug("%s waiting on condition here: ticket: %s:" % (self.name,ticket,))     
        try:
            with condition:
                condition.wait(seconds)
        except RuntimeError:
            pass
        except Exception as e:
            raise e
        
        # process finished waiting either due to time passed
        # or that resources were reserved.  
        # Hence, try to pick reserved resources
        result=self.__reserved.get(ticket, None)
        if result: 
            self.__remove_ticket(ticket)

        return result
            
    def __wait(self, sync, count, wait, callback=None):
        ''' waits for count resources, 
        
        wait uses condition object.  put method would use 
        the same condition object to notify wait of resources 
        reserved for this request.
        
        It is assumed that calling process is separated thread.  
        Otherwise, process deadlocks.
        
        Args:
            sync: (boolean) work in object synchronized, if set
            count: number of resources to wait on
            callback: callable to call back once resources are available
            
        Returns:
            list of resources, if callback is not provided (None)
            reservation ticked to use once called back, if callback is provided.
        
        '''
        seconds=None if wait <0 else wait
        condition = threading.Condition()
        caller=callback.name if hasattr(callback, 'name') else ""
        ticket=Ticket(self.name, self.__ticket_sequence())
        self.__awaiting[ticket]=(condition, count, caller,)
        
        if sync: self.__sync_release()
        
        if not callback:
            result=self.__wait_on_condition_here(condition, seconds, ticket)
            logger.debug('%s received result from condition here: %s' %(self.name, repr(result)))
        else:
            self.__wait_on_condition_callback(condition, seconds, ticket, callback)
            result=None
         
        return result

    def _get(self, sync=False, count=1, wait=-1, callback=None, hold_time=None, ticket=None):
        self.__allow_set_policy=False
        
        if ticket is not None:
            # process finished waiting either due to time passed
            # or that resources were reserved.  
            # Hence, try to pick reserved resources
            logger.debug("%s Addressing get with ticket %s" % (self.name, ticket))
            result=self.__reserved.get(ticket, None)
            if result: 
                logger.debug("%s found ticket %s" % (self.name, ticket))
                self.__remove_ticket(ticket=ticket, sync=sync)
                return result
            else:
                logger.debug("%s ticket %s not fond" % (self.name, ticket))
                     
        resource_limit=self.__policy['resource_limit']
        if resource_limit > -1 and count >resource_limit:
            raise ResourcePoolError("Trying to get count (%s) larger than resource limit (%s)" % (count, resource_limit))
        
        if sync: self.__sync_acquire()
        
        #activate_on_get=self.__policy['activate_on_get']
        # If there are awaiting processes, wait too, and this call is not after
        # put (for an awated process).
        if len(self.__awaiting) > 0 and sync:
            logger.debug("%s already has waiting list, adding this request" % (self.name,))
            resources=self.__wait(sync=sync, count=count, wait=wait, callback=callback)
            #if activate_on_get: self.__activate_allocated_resource(resources)
            return resources
        
        # try to see if request can be addressed by existing or by loading new 
        # resources.
        available_loaded=len(self.__available_resources)
        inuse_resources=len(self.__inuse_resources)
        hot_resources=available_loaded+inuse_resources
        missing_to_serve=count-available_loaded
        if missing_to_serve < 0: missing_to_serve=0
        allowed_to_load=resource_limit-hot_resources if resource_limit > 0 else missing_to_serve
        
        to_load=min(missing_to_serve, allowed_to_load)
        
        #print("TOLOAD: %s (available_loaded: %s, missing_to_serve: %s, allowed_to_load: %s, inuse_resources %s, hot_resources: %s)" % \
        #      (to_load, available_loaded, missing_to_serve, allowed_to_load, inuse_resources, hot_resources))
        
        if to_load > 0:
            logger.debug("%s needs to load additional resources: %s" % (self.name, to_load))
            self.__load(sync=False, count=to_load)
            
        # if resources are available to serve the request, do so.
        # if not, and there is wait, then do wait.
        # otherwise return no resources.
        if len(self.__available_resources) >= count:
            logger.debug("%s serving request directly: %s" % (self.name, count))
            # There are enough resources to serve!
            resources=self.__available_resources[:count]
            self.__available_resources=self.__available_resources[count:]
            self.__inuse_resources.extend(resources)
            if sync: self.__sync_release()
        elif wait != 0:
            # No resources.  But need to wait.
            logger.debug("%s cannot serve directly, waiting for availability" % (self.name))
            resources=self.__wait(sync=sync, count=count, wait=wait, callback=callback)
        else:
            # No resources and no need to wait; we are done!
            logger.debug("%s no availability and no wait" % (self.name))
            resources=[]
            if sync: self.__sync_release()
            pass
        
        return resources
    
    def get(self, count=1, wait=-1, callback=None, hold_time=None, expire=None, ticket=None):
        ''' retrieve resource from pool
        
        get checks for availability of count resources. If available: provide.
        If not available, and no wait, return empty
        If wait, wait for a seconds.  If wait < 0, wait until available.
        If about to return empty and callback is provided, register reserved with hold_time.
        When reserved, callback will be initiated with reservation_ticket.  The receiver, must call
        get(reservation_ticket=reservation_ticket) again to collect reserved resources.
        
        Important, when using wait, it is assumed that calling process is separated thread.  
        Otherwise, process deadlocks.

        Warnings: 
            1. if count > resource_limit, call will be rejected.  
            2. if count > 1 and resources are limited due to high activity or 
                clients not putting back resources, caller may be in deadlock.
            
        Args:
            count: number of resource to grab 
            wait: number of seconds to wait if none available. 
                0: don't wait
                negative: wait until available
                positive: wait period
            callback: notify that resources are available to collect
            hold_time: seconds to hold reserve resources on callback.  
                If not collected in within the specify period, reserved go
                back to pull.
            expire: seconds to limit use of resources
            ticket: reserved ticket provided in callback to allow client pick their 
                reserved resources. 
            
        Raises:
            ResourcePoolError
        '''
        
        # some validation:
        if callback and not callable(callback):
            raise ResourcePoolError("Callback must be callable, but it is no: %s" % repr(callback))
        
        result=self._get(sync=True, count=count, wait=wait, callback=callback, hold_time=hold_time, ticket=ticket)
        logger.debug("%s got result from get: %s" % (self.name, result))
        return result

    
    def __activate_allocated_resource(self, resources):
        for resource in resources:
            try:
                if not resource.active: resource.activate()
            except Exception as e:
                raise e
                   
    def put(self, *resources):
        ''' adds resource to this pool
        
        Args:
            resource: Resource object to be added to pool
        '''
        
        # validate that all resources provided are legal
        self.__sync_acquire()
        pool_resource_name=self.__resource_cls.__name__
        
        for resource in resources: 
            resource_name=resource.__class__.__name__
            if pool_resource_name != resource_name:
                raise ResourcePoolError("ResourcePool resource class (%s) doesn't match returned resource (%s)" % \
                                        (pool_resource_name, resource_name))
                
        # deposit resource back to available
        resources=list(resources)
        count=len(resources)       
        self.__available_resources.extend(self.__inuse_resources[:count])
        self.__inuse_resources=self.__inuse_resources[count:]        
        
        for ticket, (condition, count, caller) in list(self.__awaiting.items()):
            logger.debug("%s, %s found %s awaiting; require: %s, available: %s:" % (caller, self.name, ticket, count, len(self.__available_resources)))
            if count <= len(self.__available_resources):
                self.__reserved[ticket]=self._get(count=count)
                with condition:
                    logger.debug("%s, %s notifying: %s:" % (caller, self.name, ticket,))
                    condition.notify()
                del self.__awaiting[ticket]
            else:
                # this is an interesting scenario.
                # e.g., first awaiting for 3 resources. But there is only one available.
                #       second awaits for 1 resources.  If we serve it, first will have 
                #       to wait longer.  If we don't, first is holding the line.
                #       Predictive module will learn if it is better to hold the line,
                #       
                break
        self.__sync_release()      


class RequestorCallback(object):
    def __init__(self, notify_queue):
        self.q=notify_queue
    def __call__(self, ticket=None):
        logger.debug('RequestorCallback notifying ticket: %s' %(ticket,))
        self.q.put(ticket)

class Requestor(object):
    ''' Manages a single request from multiple resource pools
    '''
        
    def __init__(self, request, wait=-1, callback=None, hold_time=None, expire=None, audit=True):
        '''Initialize requestor object to manage request from multiple pools
        
        Args:
            request: iterator on tuples of resource pool object and quantity of resources required
            wait: seconds to wait for resources
            callback: callable to callback when resources are reserved
            hold_time: how long to hold resources in reserved
            expire: seconds to limit use of resources.
            audit: if True, ensures resources returened are from the same requestor 
            
        '''
        self.__request=dict([(r.name, (r,count)) for r, count in request])
        self.__wait=wait
        self.__callback=callback
        self.__hold_time=hold_time
        self.__notify_queue=queue.Queue()
        self.__tickets=list()
        self.__resources=dict()
        self.__reserved=False
        self.__resource_pool_requestor_lock=threading.Lock()
        self.__collect_resources_started=False
        self.__audit=audit
        if callback and hasattr(callback, 'name'):
            self.__client_name=callback.name
        else: self.__client_name=''
        self.__get()
    
    def __sync_acquire(self):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        logger.debug("ResourcePoolRequestor Acquiring Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_requestor_lock.acquire()
        
    def __sync_release(self):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        logger.debug("ResourcePoolRequestor Releasing Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_requestor_lock.release()

    def __is_reserved(self):
        ''' Returns True if all resources are collected
        '''
        missings=set(self.__request.keys()) - set(self.__resources.keys())
        self.__reserved=len(missings) ==0
        #if not self.__reserved:
        #    logger.debug("%s missing resources %s" %(self.__client_name, missings))
        return self.__reserved
        
    def is_reserved(self):
        ''' Returns True if all resources are collected
        '''
        return self.__reserved
        
    def __get(self):
        callback=RequestorCallback(self.__notify_queue) if self.__callback else None
        for rp, count in self.__request.values():
            logger.debug("%s requesting resources %s(%s)" %(self.__client_name, rp.name, count))
            response=rp.get(count=count, wait=self.__wait, callback=callback, hold_time=self.__hold_time)
            
            if response:
                logger.debug("%s received resources %s" %(self.__client_name, response))
                self.__resources[rp.name]=response
                
        if not self.__is_reserved():
            go_collect=False
            self.__sync_acquire()
            if not self.__collect_resources_started:
                go_collect=True
                self.__collect_resources_started=True
            self.__sync_release()
            if go_collect:
                logger.debug("%s going to collect missing resources" % (self.__client_name, ))
                self.__collect_resources()   
        else:
            self.__notify_collected()
    
    @Threaded()
    def __collect_resources(self):
        start_time=time.time()
        go=True
        wait=self.__wait if self.__wait is None or self.__wait >=0 else None
        while go:
            logger.debug("%s trying to get from notify queue (wait: %s)" %(self.__client_name, wait))
            try:
                ticket=self.__notify_queue.get(timeout=wait)
            except queue.Empty:
                ticket=None
            
            if ticket is not None:
                rp_name=ticket.pool_name
                rp, _=self.__request[rp_name]
                resources=rp.get(ticket=ticket)
                logger.debug("Collected resources %s" %(resources))
                self.__resources[rp_name]=resources
                #self.__resources[rp_name]=dict([(r.id_, r) for r in resources])
                
            time_passed=time.time() - start_time
            if self.__wait and self.__wait >0:
                wait=self.__wait - time_passed
                go=wait > 0
            logger.debug("%s all reserved %s" % (self.__client_name, self.__is_reserved(),))
            go=go and not self.__is_reserved()
        
        if self.__is_reserved():
            self.__notify_collected()
        else:
            logger.debug("%s couldn't collect all resources, releasing collected." % (self.__client_name,))
            for rp_name, resources in list(self.__resources.items()):
                rp, _=self.__request[rp_name]
                #returning=list(resources.values())
                rp.put(*resources)
                del self.__resources[rp_name]              
                 
    def __notify_collected(self):
        if self.__is_reserved() and self.__callback:
            self.__callback(True)
                  
    def get(self,):
        result=None
        if self.__is_reserved():
            result=list()
            for k, r in list(self.__resources.items()): 
                result.extend(r) 
                #if not self.__audit: del self.__resources[k]
             
        return result 
            
    def put(self, *resources):
        logger.debug("%s putting back resources %s" % (self.__client_name, repr(resources)))
        
        # ensure resources returned are in alignment with this container.
        #resource_to_return=list()
        resource_count=dict()
        for resource in resources:
            rp_name=resource.pool 
            try:
                count=resource_count[rp_name]
            except KeyError:
                count=list()
                resource_count[rp_name]=count
            count.append(resource)
            
        for rp_name, count in list(resource_count.items()):
            logger.debug("%s ResourcePool %s removing %s resources" % (self.__client_name, rp_name, count))
            rp_resources=self.__resources[rp_name]
            self.__resources[rp_name]=rp_resources[len(count):]
            rp, _ = self.__request[rp_name]
            rp.put(*count)    
            
    def put_requested(self, request):
        self.__sync_acquire()
        logger.debug("putting back request: %s" % (repr(request)))
        
        # ensure resources returned are in alignment with this container.
        #resource_to_return=list()
        for rp, count in request:
            logger.debug("ResourcePool %s removing %s resources" % (rp.name, count))
            resources=self.__resources[rp.name]
            #pool=self.__pools[rp.name]
            to_return=resources[:count]
            rp.put(*to_return)
            del resources[:count]
        
        self.__sync_release()
                    
    def __del__(self):
        # need to make sure all resources are returned
        for rp_name, resources in list(self.__resources.items()):
            rp, _ = self.__request[rp_name]
            rp.put(*resources)
                
        # TODO: notify ResourcePool that not waiting on resources anymore.


class RequestorsCallback(object):
    def __init__(self, notify_queue, request_id):
        self.q=notify_queue
        self.request_id=request_id
    def __call__(self, ticket=None):
        logger.debug('RequestorsCallback notifying request id %s ticket: %s' % (self.request_id, ticket, ))
        self.q.put( (ticket, self.request_id,) )
         
Pool=namedtuple('Pool', ['pool', 'reserved', 'inuse'])

class Requestors(object):
    ''' Manages multiple requests from multiple resource pools
    '''
    
    __request_id=Sequence("Requestors_request_id")
    
    class Request(object):
        def __init__(self, request_id, request, wait=-1, callback=None, hold_time=None, expire=None):
            self.request_id=request_id
            self.request=dict([(r.name, (r,count)) for r, count in request])
            self.wait=wait
            self.callback=callback
            self.hold_time=hold_time
            self.reserved=False
            self.fetched=False
            self.resources=dict()
            self.start_time=time.time()
            if callback and hasattr(callback, 'name'):
                self.client_name=callback.name
            else: self.client_name=''
            
        def __repr__(self):
            return "%s" %(self.request_id, )
        
    def __init__(self, audit=False):
        self.__audit=audit
        self.__pools=dict() # mapping for pool names to namedtuple of pool, requestor, and resources
        self.__requests=dict()
        self.__notify_queue=queue.Queue()
        self.__resource_pool_requestor_lock=threading.Lock()
        self.__collect_resources_started=False
        
    def reserve(self, request, wait=-1, callback=None, hold_time=None, expire=None,):
        '''Initialize requestor object to manage request from multiple pools
        
        Args:
            request: iterator on tuples of resource pool object and quantity of resources required
            wait: seconds to wait for resources
            callback: callable to callback when resources are reserved
            hold_time: how long to hold resources in reserved
            expire: seconds to limit use of resources.
            audit: if True, ensures resources returned are from the same requestor 
            
        '''
        request_id=self.__request_id()
        self.__requests[request_id]=self.Request(request_id=request_id, request=request, wait=wait, callback=callback, hold_time=hold_time, expire=expire)        
        self.__get(request_id)
        return request_id
    
    def __sync_acquire(self):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        logger.debug("ResourcePoolRequestors Acquiring Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_requestor_lock.acquire()
        
    def __sync_release(self):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        logger.debug("ResourcePoolRequestors Releasing Lock; %s.%s(%s)" % (filename, function_name, line_number))
        self.__resource_pool_requestor_lock.release()

    def __is_reserved(self, request_id):
        ''' Returns True if all resources are collected
        '''
        request=self.__get_request(request_id)
        
        missings=set(request.request.keys()) - set(request.resources.keys())
        request.reserved=len(missings) ==0
        if not request.reserved:
            logger.debug("%s missing resources %s" %(request.client_name, missings))
        return request.reserved
        
    def is_reserved(self, request_id):
        ''' Returns True if all resources are collected
        '''
        try:
            result=self.__requests[request_id].reserved
        except:
            raise ResourcePoolError("Unknown request_id: %s" % request_id)
        return result
        
    def __get(self, request_id):
        request=self.__get_request(request_id)
        callback=RequestorsCallback(self.__notify_queue, request_id=request_id) if request.callback else None
        for rp, count in request.request.values():
            logger.debug("%s requesting resources %s(%s)" %(request.client_name, rp.name, count))
            resources=rp.get(count=count, wait=request.wait, callback=callback, hold_time=request.hold_time)
            
            if resources:
                logger.debug("%s received resources %s" %(request.client_name, resources))
                #self.__sync_acquire()
                self.__update_resource_store(request_id=request_id, resource_pool=rp, resources=resources)
                #self.__sync_release()
                
        if not self.__is_reserved(request_id):
            go_collect=False
            #self.__sync_acquire()
            logger.debug("%s __collect_resources_started %s" %(request.client_name, self.__collect_resources_started))
            if not self.__collect_resources_started:
                go_collect=True
                self.__collect_resources_started=True
            #self.__sync_release()
            if go_collect:
                logger.debug("%s starting collector" % (request.client_name, ))
                self.__collect_resources()   
        else:
            self.__sync_acquire()
            self.__notify_collected(request_id)
            self.__sync_release()
            
    def __return_resources(self, request_id):
        logger.debug("returning resources  for request id: %s" %(request_id,))
        request=self.__get_request(request_id)
        for rp_name, resources in list(request.resources.items()):
            rp=self.__pools[rp_name]
            returning=list(resources.values())
            rp.put(*returning)
            del request.resources[rp_name]               
    
    def __update_resource_store(self, request_id, resource_pool, resources):
        logger.debug("Collected resources %s" %(resources))
        request=self.__get_request(request_id)
        #new_resources=[r.setattrib('requestor_request_id', request_id)for r in resources]
        request.resources.update({resource_pool.name: len(resources),})
        # plasce resources in reserved
        #self.__pools[resource_pool.name]=resource_pool
        
        try:
            pool=self.__pools[resource_pool.name]
        except KeyError:
            pool=Pool(resource_pool, list(), list())
            self.__pools[resource_pool.name]=pool
        pool.reserved.extend(resources)
        #rp_resources.extend(map(lambda x: x.setattrib('__requestors_request_id', request_id), resources))
        
        return self.__is_reserved(request_id)
    
    def __get_request(self, request_id, default=None):
        request=self.__requests.get(request_id, None)
        if request is None and default is None:
            (frame, filename, line_number,
             function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]            
            raise RequestNotFound("Unknown request_id: %s: %s(%s)" % (request_id, function_name, line_number,))
        elif request is None:
            request=default
        return request
    
    def __evaluate_collect_wait_time(self):
        waits=list() 
        found_endless_waiter=False 
        for request_id, request in list(self.__requests.items()):
            if self.__is_reserved(request_id): continue
            time_passed=time.time() - request.start_time
            if request.wait and request.wait >0:
                wait=request.wait - time_passed
                if wait<= 0: 
                    self.__return_resources(request_id)
                else: waits.append(wait)
            else:
                found_endless_waiter=True
        self.__collect_resources_started=len(waits)>0 or found_endless_waiter
        len_waits=len(waits)
        wait=min(waits) if len_waits>0 else None  
        logger.debug("Wait evaluated: %s: waits: %s: found endless: %s" %(wait, len(waits), found_endless_waiter,))
        return wait
    
    @Threaded()
    def __collect_resources(self):
        wait=self.__evaluate_collect_wait_time()
        logger.debug("starting get loop (wait: %s, collect: %s)" %(wait,self.__collect_resources_started))
        #self.__collect_resources_started=True
        while self.__collect_resources_started:
            logger.debug("trying to get from notify queue (wait: %s)" %(wait,))
            try:
                ticket, request_id=self.__notify_queue.get(timeout=wait)
            except queue.Empty:
                ticket, request_id=None, None
            
            logger.debug("got something from notify queue request_id: %s, ticket: %s" %(request_id, ticket,))
            
            self.__sync_acquire()
            
            if request_id is not None:
                rp_name=ticket.pool_name
                request=self.__get_request(request_id)
                logger.debug("%s received resources for request_id %s" %(request.client_name, request_id,))
                rp, _=request.request[rp_name]
                resources=rp.get(ticket=ticket)
                reserved=self.__update_resource_store(request_id=request_id, resource_pool=rp, resources=resources)
                if reserved:
                    self.__notify_collected(request_id)          
            
            wait=self.__evaluate_collect_wait_time()
            
            self.__sync_release()
                    
    def __notify_collected(self, request_id):
        request=self.__get_request(request_id)
        if self.__is_reserved(request_id) and request.callback:
            logger.debug("%s Calling callback for %s" %(request.client_name, request_id,))
            request.callback(request_id)
                  
    def was_fetched(self, request_id):
        request=self.__get_request(request_id)
        return request.fetched
                  
    def get(self, request_id):
        result=None
        request=self.__get_request(request_id)
        self.__sync_acquire()
        
        if request.reserved and not request.fetched:
            result=list()
            #for _, resources in list(request.resources.items()): 
            for rp_name, count in list(request.resources.items()):
                pool=self.__pools[rp_name]
                resources=pool.reserved[:count]
                del pool.reserved[count:]
                pool.inuse.extend(resources)
                result.extend(resources) 
            request.fetched=True    
            request.request.clear()
            logger.debug("%s remove required and resources %s" %(request.client_name, request_id,))
        
        self.__sync_release() 
        return result 
            
    def put(self, *resources):
        #self.__sync_acquire()
        logger.debug("putting back resources %s" % (repr(resources)))
        pool_count=dict()
        for resource in resources:
            rp_name=resource.pool
            try: 
                count=pool_count.get[rp_name]
            except KeyError:
                count=list()
                pool_count.get[rp_name]=list
            count.append(resource)
        
        # ensure resources returned are in alignment with this container.
        self.put_requested([ (self.__pools[rp_name].pool, count) 
                            for rp_name, count in list(pool_count.items())])    

    def put_requested(self, request):
        ''' returns fetched request formated as in get.
        
        Args:
            request: list of tuples of resource pool and amount requested
        '''
        
        self.__sync_acquire()
        logger.debug("putting back request: %s" % (repr(request)))
        
        # ensure resources returned are in alignment with this container.
        #resource_to_return=list()
        for rp, count in request:
            logger.debug("ResourcePool %s removing %s resources" % (rp.name, count))
            pool=self.__pools[rp.name]
            to_return=pool.inuse[:count]
            pool.pool.put(*to_return)
            del pool.inuse[:count]
        
        self.__sync_release()
        
    # TODO: rebuild del of captured resources
    def __del__(self):
        # need to make sure all resources are returned
        for pool in self.__pools.values():
            for resource in pool.inuse + pool.reserved:
                pool.pool.put(resource)




'''
Created on Oct 31, 2016

@author: arnon
'''

import logging

module_logger=logging.getLogger(__name__)


class Delay(object):
    """Event object used to reuse in user program.
        
        Event objects are created by Eventor.add_event programming interface or by the mechanism itself
    
        Attributes:
            N/A
            
        Methods:
            db_write: write event to db file
    
    """

    def __init__(self, delay_id, func, seconds, event):
        self.func=func
        self.seconds=seconds
        self.delay_id=delay_id #get_event_id(
        self.event=event
                    
    def __repr__(self):
        return "Delay(%s, %s)" % (self.delay_id, self.seconds)
        
    def __str__(self):
        return repr(self)
    
    #def db_write(self, db):
    #    delay=db.add_delay(delay_id=self.delay_id, seconds=self.seconds, active=self.active, activated=self.activated, )
    #    self.db_delay=delay
    
if __name__ == '__main__':
    e1=Delay('E1')
    e2=Delay('E2')
    e3=Delay('E3')
    e4=Delay('E4')
    


'''
Created on Oct 21, 2016

@author: arnon
'''

from sqlalchemy import Column, UniqueConstraint, Sequence, create_engine
from sqlalchemy import Integer, BigInteger, String, PickleType, DateTime, Boolean
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime
from sqlalchemy import Enum as SQLEnum

from eventor.eventor_types import TaskStatus
from namedlist import namedlist


Info=namedlist("InfoRow", "id_ run_id name value")


def info_from_db(tobj):
    result=Info(tobj.name, tobj.value)
    return result


def info_to_db(obj, tcls):
    result=tcls(obj.name, obj.value)
    return result


def info_table(base):
    class Info(base):
        __tablename__ = 'Info'
        
        id_ = Column(Integer, Sequence('trigger_id_seq'), primary_key=True)
        run_id=Column(String, nullable=False)
        name=Column(String, nullable=False)
        value=Column(String, nullable=True)
        
        __table_args__ = (
                UniqueConstraint('run_id', 'name'),
                )
    
        def __repr__(self):
            return "<Info(run_id='%s', name='%s', value='%s')>" % (self.run_id, self.name, self.value)
        
    return Info


Trigger=namedlist("Trigger", "id_ run_id event_id sequence recovery created acted")


def trigger_from_db(tobj):
    result=Trigger(tobj.id_, tobj.run_id, tobj.event_id, tobj.sequence, tobj.recovery, tobj.created, tobj.acted)
    return result


def trigger_to_db(obj, tcls):
    result=tcls(obj.id_, obj.run_id, obj.eveny_id, obj.sequence, obj.recovery, obj.created, obj.acted)
    return result


def trigger_table(base):
    class Trigger(base):
        __tablename__ = 'Trigger'
        
        id_ = Column(Integer, Sequence('trigger_id_seq'), primary_key=True)
        run_id = Column(String, nullable=False, default='')
        event_id = Column(String, nullable=False)
        sequence = Column(Integer, nullable=False)
        recovery = Column(Integer, nullable=False, default=0)
        created = Column(DateTime(), default=datetime.utcnow) 
        acted = Column(DateTime(), nullable=True) 
    
        def __repr__(self):
            return "<Trigger(id='%s', run_id='%s', event_id='%s', sequence='%s', recovery='%s', created='%s', acted='%s')>" % (
                self.id_, self.run_id, self.event_id, self.sequence, self.recovery, self.created, self.acted)
    
    return Trigger


Task=namedlist("Task", "id_ run_id step_id sequence recovery pid status result created updated") 

       
def task_from_db(tobj):
    result=Task(tobj.id_, tobj.run_id, tobj.step_id, tobj.sequence, tobj.recovery, tobj.pid, tobj.status, tobj.result, tobj.created, tobj.updated)
    return result


def task_to_db(obj, tcls):
    result=tcls(obj.id_, obj.run_id, obj.step_id, obj.sequence, obj.recovery, obj.pid, obj.status, obj.result, obj.created, obj.updated)
    return result


def task_table(base):
    class Task(base):
        __tablename__ = 'Task'
        
        id_ = Column(Integer, Sequence('task_id_seq'), primary_key=True)
        run_id = Column(String, default='')
        step_id = Column(String,)
        sequence = Column(Integer,)
        recovery = Column(Integer, nullable=False)
        pid = Column(Integer, nullable=True)
        status = Column(SQLEnum(TaskStatus), ) 
        result = Column(PickleType() , nullable=True,)
        created = Column(DateTime(), default=datetime.utcnow) 
        updated = Column(DateTime(), nullable=True, ) 
        
        __table_args__ = (
                UniqueConstraint('run_id', 'step_id', 'sequence', 'recovery'),
                )
    
        def __repr__(self):
            return "<Task(id='%s', run_id='%s', step_id='%s', sequence='%s', recovery='%s', pid='%s', status='%s', created='%s', updated='%s')>" % (
                self.id_, self.run_id, self.step_id, self.sequence, self.recovery, self.pid, self.status, self.created, self.updated)
    
    return Task


Delay=namedlist("Delay", "id_ run_id delay_id sequence recovery seconds active activated updated")


def delay_from_db(tobj):
    result = Delay(tobj.id_, tobj.run_id, tobj.delay_id, tobj.sequence, tobj.recovery, tobj.seconds, tobj.active, tobj.activated, tobj.updated)
    return result


def delay_to_db(obj, tcls):
    result = tcls(obj.id_, obj.run_id, obj.delay_id, obj.sequence, obj.recovery, obj.active, obj.activated, obj.updated)
    return result


def delay_table(base):
    class Delay(base):
        __tablename__ = 'Delay'
        
        id_=Column(Integer, Sequence('delay_id_seq'), primary_key=True)
        run_id = Column(String, default='')
        delay_id = Column(String,)
        sequence = Column(Integer,)
        recovery = Column(Integer,)
        seconds = Column(BigInteger, nullable=True)
        active = Column(Boolean, default=False)
        activated = Column(DateTime(), nullable=True,) 
        updated = Column(DateTime(), default=datetime.utcnow) 
        
        __table_args__ = (
                UniqueConstraint('run_id', 'delay_id', 'sequence', 'recovery'),
                )
    
        def __repr__(self):
            return "<Delay(id='%s', run_id='%s', delay_id='%s', delay='%s', active='%s', activated='%s')>" % (
                self.id_, self.run_id, self.delay_id, self.seconds, self.active, self.activated)
    
    return Delay   
 
        
if __name__ == '__main__':
    #engine = create_engine('sqlite:///:memory:', echo=True)
    file='/var/acrisel/sand/eventor/eventor/eventor/eventor/schema.db'
    dns='sqlite:///' + file
    try:
        import os
        os.remove(file)
    except FileNotFoundError:
        pass
    
    Base=declarative_base()
    engine = create_engine(dns, echo=True)
    Info=info_table(Base)
    Trigger=trigger_table(Base)
    Task=task_table(Base)
    Delay=delay_table(Base)
    Base.metadata.create_all(engine)
    
    '''
    from sqlalchemy import MetaData
    from sqlalchemy_schemadisplay import create_schema_graph
    
    graph = create_schema_graph(
      metadata = MetaData(dns),
      show_datatypes=True, # The image would get nasty big if we'd show the datatypes
      show_indexes=True, # ditto for indexes
      #rankdir='LR', # From left to right (instead of top to bottom)
      concentrate=True # Don't try to join the relation lines together
    )

    graph.write_png('schema.png')
    '''
    
    from sqlalchemy.orm import sessionmaker
    
    Session = sessionmaker(bind=engine)
    session = Session()
    
    db_task=Task(step_id='34',status=TaskStatus.ready , recovery='0', sequence='45')
    session.add(db_task)
    session.flush() 
    session.commit() 
    
    db_trigger=Trigger(event_id='46' , sequence='45')
    session.add(db_trigger)
    session.flush() 
    session.commit() 

    db_info=Info(run_id='123', name='46' , value='45')
    session.add(db_info)
    session.flush() 
    session.commit() 

    db_delay=Delay(run_id='123', delay_id='S1', recovery=3 )
    session.add(db_delay)
    session.flush() 
    session.commit() 

    db_delay=Delay(delay_id='mydelay', seconds=2419200, sequence=3, )
    print('DBAPI - add_delay: %s' % (repr(db_delay), ))
    session.add(db_delay)
    session.flush() 
    session.commit() 

    


'''
Created on Jun 20, 2017

@author: arnon
'''

import os
import yaml
import logging
from eventor.expandvars import expandvars

logger=logging.getLogger(__file__)


from collections import Mapping

def keys2lower(map_):
    ''' Build the same dict but with keys in lower case
    '''
    if not isinstance(map_, Mapping): return map_
    
    type_=type(map_)
    nmap=type_([(k.lower(), keys2lower(v)) for k,v in map_.items()])
    return nmap

def expandmap(map_, expand_map=None):
    ''' Build the same dict but with keys in lower case
    '''
    if not expand_map:
        expand_map=os.environ
    
    type_=type(map_)
    nmap=type_([(k, expandvars(v, expand_map)) for k,v in map_.items()])
    return nmap

def getconfdict(conf=None, root='', expand=True, expand_map=None):
    result=conf
    if root is not None:
        for key in root.split('.'):
            result=result.get(key, result.get(key.swapcase(), result))
    if expand:
        result=expandmap(result, expand_map=expand_map)
    
    return result


def getconfstr(conf=None, root=None, expand=True, expand_map=None):
    ''' YAML text or name of file.
    '''
    
    # try first as file
    path=os.path.abspath(conf)
    if os.path.isfile(path):
        with open(path, 'r') as f:
            text=f.read()
    else:
        text=conf
        
    content=yaml.load(text)
    result=getconfdict(conf=content, root=root, expand=expand, expand_map=expand_map)
    return result


def getconffile(conf=None, root=None, expand=True, expand_map=None):
    '''file object opened to read: read file text and treat as str.
    '''
    text=conf.read()
    return getconfstr(conf=text, root=root, expand=expand, expand_map=expand_map)

type_map={
    'dict': getconfdict,
    'MergedChainedDict': getconfdict,
    'str': getconfstr,
    'file': getconffile,
    }


def getrootconf(conf=None, root=None, expand=True, expand_map=None):
    '''Translate configuration into sqlalchemy url
    
    Args:
        conf: dictionary, text, filename, or 
    '''
    
    type_=type(conf).__name__
    try:
        func=type_map.get(type_)
    except Exception as e:
        raise Exception('unhandled conf type: %s, allow %s.' %(type_, repr(list(type_map.keys())))) from e
    
    rootconf=func(conf=conf, root=root, expand=expand, expand_map=expand_map)
    
    return rootconf    

if __name__ == '__main__':
    from pprint import pprint
    pprint(getrootconf('db.conf', 'default'))
    pprint(getrootconf(conf='db.conf', root='playpg'))
    pprint(getrootconf(conf='db.conf', root='playmem'))
    pprint(getrootconf(conf='db.conf', root='playfile'))
    pprint(getrootconf('db.conf', 'playmy'))

    


'''
Created on Oct 21, 2016

@author: arnon
'''

from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
import logging
import os
import multiprocessing 
import threading
from datetime import datetime

from eventor.dbschema import task_table, trigger_table, delay_table, info_table
from eventor.dbschema import trigger_from_db, task_from_db, delay_from_db
from eventor.eventor_types import DbMode, Invoke
from eventor.utils import decorate_all, print_method, calling_module
from eventor.sqladburl import SQLAlchemyConf

module_logger=logging.getLogger(__name__)


class DbApiError(Exception): pass
    
        
def get_sqlalchemy_conf(modulefile, userstore, config, root=None, echo=False):
    '''
        If userstore is provided:
            If it is a database key in config:
                use that database
            elif ':memory:':
                use in-memory
            else:
                use as runfile
        elif database has default key:
            use default database
        elif modulefile  is provided:
            use modulefile
        else:
            use memory
    '''
    
    if not root:
        root=os.environ.get('EVENTOR_DB_CONFIG_TAG', 'DATABASES')
        
    database=userstore if userstore else 'default'
    try:
        sqldb=SQLAlchemyConf(conf=config, root=root, database=database, quiet=False, echo=echo)
    except Exception:
        if userstore==':memory:':
            conf={'DB': {'adhoc': {'dialect': 'sqlite', 'query': {'cache':'shared'}}}}
        elif userstore:
            conf={'DB': {'adhoc': {'dialect': 'sqlite', 'database':userstore,}}} # 'query': {'cache':'shared'}}}}
        elif modulefile:
            conf={'DB': {'adhoc': {'dialect': 'sqlite', 'database':modulefile,}}} # 'query': {'cache':'shared'}}}}
        else:
            conf={'DB': {'adhoc': {'dialect': 'sqlite', 'query': {'cache':'shared'}}}}
            
        module_logger.debug("Using adhoc DB config: %s" %(repr(conf)))
        sqldb=SQLAlchemyConf(conf=conf, root='DB', database='adhoc',quiet=True, echo=echo)
    else:
        module_logger.debug("Using DB config: %s" %(repr(config)))
    module_logger.debug("SQLAlchemyConf: %s" % (repr(sqldb),))
    return sqldb
 
class DbApi(object):

    def __init__(self, modulefile=None, shared_db=False, run_id='', userstore=None, session=None, mode=DbMode.write, config={}, root=None, thread_sync=False, echo=False, logger=None):
        '''
        Args:
            moduefile (path): file in which Eventor data would be stored and managed for reply/restart 
            if the value is :memory:, an in-memory temporary structures will be used
            userstore (path): store file provided by user
            
        Devising storage:
            If Store is provided:
                If it is a database key in config:
                    use that database
                else:
                    use as runfile
            else:
                if database has default key:
                    use default database
                else:
                    use modulefile if provided
        

        '''
        global module_logger
        if logger:
            module_logger = logger
        
        self.engine = None
        self.session = None
        self.run_id = run_id
        self.shared_db = shared_db
        self.thread_sync = thread_sync
        
        self.__sqlalchemy = get_sqlalchemy_conf(modulefile, userstore, config, root=root, echo =echo)
        self.metadata = metadata=self.__sqlalchemy.get_metadata()
        
        Base = declarative_base(metadata=metadata)
        self.Info = info_table(Base)
        self.Task = task_table(Base)
        self.Delay = delay_table(Base)
        self.Trigger = trigger_table(Base)

        if session:
            self.session = session()   
        else: 
            self.session=self.__sqlalchemy.get_session()
        self.open(mode=mode) 
            
        self.db_transaction_lock=threading.Lock()
    
    def set_thread_synchronization(self, value=True):
        self.thread_sync=value
        
    def lock(self):
        if self.thread_sync:
            self.db_transaction_lock.acquire()
            
    def release(self):
        if self.thread_sync:
            self.db_transaction_lock.release()
    
    def open(self, mode=DbMode.write):        
        if mode == DbMode.write:
            module_logger.debug("DbMode write: creating schema")
            self.create_db() 
            
    def close(self):
        self.session.remove()
    
    def create_schema(self):
        metadata = self.metadata
        schemas = set()
        for table in metadata.tables.values():
            if table.schema is not None:
                schemas.add(table.schema)
        
        if len(schemas) > 0:
            for schema in schemas:
                if not self.shared_db:
                    module_logger.debug("Not in shared_db mode: dropping before creating schema %s" % (schema,))
                    metadata.bind.execute("DROP SCHEMA IF EXISTS %s CASCADE " % schema)
                metadata.bind.execute("CREATE SCHEMA IF NOT EXISTS %s" % schema)
        elif self.__sqlalchemy.dbconf['dialect'] == 'sqlite':
            # in this case we need to remove database
            if not self.shared_db:
                database=self.__sqlalchemy.dbconf['database']
                if os.path.isfile(database):
                    module_logger.debug("Not in shared_db mode: removing database file: %s" % (database))
                    os.remove(database)

        
    def create_db(self, ):
        #self.__create_engine(runfile)
        self.create_schema()
        self.metadata.create_all()
        
    def commit_db(self):
        self.session.flush()
        self.session.commit()
           
    def write_info(self, **info):
        self.lock()
        for name, value in info.items():
            db_info=self.Info(run_id=self.run_id, name=name, value=value)
            self.session.add(db_info)
            self.commit_db()
        self.release()
        
    def read_info(self, ):
        self.lock()
        rows = self.session.query(self.Info).filter(self.Info.run_id==self.run_id).all()
        self.release()
        info=dict()
        for row in rows:
            info[row.name]=row.value
        return info
        
    def update_info(self, **info):
        self.lock()
        for name, value in info.items():
            self.session.query(self.Info).filter(self.Info.run_id==self.run_id, self.Info.name==name).update({self.Info.name:name, self.Info.value:value}, synchronize_session=False)
        self.release()
                
    def get_trigger_iter(self, recovery):
        self.lock()
        rows = self.session.query(self.Trigger).filter(self.Trigger.run_id==self.run_id, self.Trigger.recovery==recovery).all()
        #rows = query.statement.execute().fetchall()
        self.release()
        for row in rows:
            yield trigger_from_db(row)
        #return rows
    
    def get_trigger_map(self, recovery=0):
        self.lock()
        triggers = self.session.query(self.Trigger).filter(self.Trigger.run_id==self.run_id, self.Trigger.recovery==recovery).all()
        self.release()
        trigger_map=dict()
        for trigger in triggers:
            try:
                sequence_map=trigger_map[trigger.sequence]
            except KeyError:
                sequence_map=dict()
                trigger_map[trigger.sequence]=sequence_map
            sequence_map[trigger.event_id]=trigger_from_db(trigger)
        return trigger_map
        
    def add_trigger(self, event_id, sequence, recovery):
        trigger = self.Trigger(run_id=self.run_id, event_id=event_id, sequence=sequence, recovery=recovery)
        self.lock()
        self.session.add(trigger)
        self.commit_db()
        self.release()
        return trigger_from_db(trigger)
        
    def add_trigger_if_not_exists(self, event_id, sequence, recovery):
        self.lock()
        module_logger.debug("DBAPI - cehcking if event trigger do not exist: %s(%s)" %(event_id, sequence))
        trigger = self.session.query(self.Trigger).filter(self.Trigger.run_id==self.run_id, self.Trigger.event_id==event_id, self.Trigger.sequence==sequence, self.Trigger.recovery==recovery)
        found = self. session.query(trigger.exists()).scalar()
        if not found:
            module_logger.debug("DBAPI - adding event trigger %s(%s)" %(event_id, sequence))
            trigger = self.Trigger(run_id=self.run_id, event_id=event_id, sequence=sequence, recovery=recovery)
            self.session.add(trigger)
            self.commit_db()
        self.release()
        return trigger_from_db(trigger)
    
    def _get_trigger(self, trigger):
        rows = self.session.query(self.Trigger).filter(self.Trigger.id_==trigger.id_).all()
        try: return rows[0]
        except: return None
    
    def acted_trigger(self, trigger):
        self.lock()
        db_trigger=self._get_trigger(trigger)
        trigger.acted=db_trigger.acted=datetime.utcnow()
        #self.session.add(trigger)
        self.commit_db()
        self.release()
        return trigger
    
    def count_trigger_ready(self, sequence=None, recovery=None ):
        self.lock()
        members=self.session.query(self.Trigger).filter(self.Trigger.run_id == self.run_id, self.Trigger.acted == None, self.Trigger.recovery == recovery)
        if sequence:
            members=members.filter(self.Trigger.sequence==sequence)
        count=members.count()
        self.release()
        return count
        
    def count_trigger_ready_like(self, sequence, recovery):
        self.lock()
        try:
            members=self.session.query(self.Trigger).filter(self.Trigger.sequence.like(sequence), self.Trigger.run_id == self.run_id, self.Trigger.acted == None, self.Trigger.recovery==recovery)
        except:
            self.release()
            raise
        count=members.count()
        self.release()
        return count
        
    def add_task(self, step_id, sequence, status, recovery=None):
        self.lock()
        task=self.Task(run_id=self.run_id, step_id=step_id, sequence=sequence, status=status, recovery=recovery)
        self.session.add(task)
        self.commit_db()
        self.release()
        return task_from_db(task)
        
    def add_task_if_not_exists(self, step_id, sequence, status, recovery=None):
        self.lock()
        task=self.session.query(self.Task).filter(self.Task.run_id==self.run_id, self.Task.sequence==sequence, self.Task.step_id == step_id, self.Task.recovery==recovery)
        found = self. session.query(task.exists()).scalar()
        if not found:
            task=self.Task(run_id=self.run_id, step_id=step_id, sequence=sequence, status=status, recovery=recovery)
            self.session.add(task)
            self.commit_db()
        self.release()
        result=task_from_db(task)
        module_logger.debug('DBAPI - add_task_if_not_exists: %s' % (repr(result), ))
        return result
        
    def update_task(self, task, session=None):
        self.lock()
        if not session:
            session=self.session
        task.updated=updated=datetime.utcnow()
        updates={self.Task.status:task.status, self.Task.updated: updated,}
        if task.pid:
            updates[self.Task.pid]=task.pid                            
        if task.result:
            updates[self.Task.result]=task.result                                                                   
        self.session.query(self.Task).filter(self.Task.id_==task.id_).update(updates, synchronize_session=False)
        self.commit_db()
        self.release()
        module_logger.debug('DBAPI - update_task: %s' % (repr(task), ))
        return task
        
    def update_task_status(self, task, status, session=None,):
        self.lock()
        #if isinstance(task, int):
        #    task_id=task
        #elif isinstance(task, Task):
        #    task_id=task.id_
        #else:
        #    raise DbApiError("Unknown task type (%s), expected int or Task" % (type(task), ))
        
        task_id=task.id_
        
        if not session:
            session=self.session
        task.updated=updated=datetime.utcnow()
        task.status=status
        module_logger.debug('updating task status: %s(%s)' % (task_id, status))
        updates={self.Task.status: status, self.Task.updated: updated,}                                               
        self.session.query(self.Task).filter(self.Task.id_==task_id).update(updates, synchronize_session=False)
        self.commit_db()
        self.release()
        return task
        
    def get_task_iter(self, recovery, status=None):
        # TODO: do we really need recovery here
        self.lock()
        rows = self.session.query(self.Task)
        if status:
            rows = self.session.query(self.Task).filter(self.Task.run_id==self.run_id, self.Task.status.in_(status)).all()
        else:
            rows = self.session.query(self.Task).filter(self.Task.run_id==self.run_id).all()
        self.release()
        for row in rows:
            result=task_from_db(row)
            module_logger.debug("task_iter: task: %s" %(repr(result)))
            yield  result
    
    def get_task_map(self, recovery=0):
        self.lock()
        tasks = self.session.query(self.Task).filter(self.Task.run_id==self.run_id, self.Task.recovery==recovery).all()
        task_map=dict()
        for task in tasks:
            try:
                sequence_map=task_map[task.sequence]
            except KeyError:
                sequence_map=dict()
                task_map[task.sequence]=sequence_map
            sequence_map[task.step_id]=task_from_db(task)
        self.release()
        module_logger.debug("get_task_map: task: %s" %(repr(task_map)))
        return task_map
    
    def count_tasks(self, recovery, status, sequence=None):
        self.lock()
        with self.session.no_autoflush:
            members=self.session.query(self.Task)
        members=members.filter(self.Task.run_id==self.run_id, self.Task.status.in_(status), self.Task.recovery==recovery)
        if sequence:
            members=members.filter(self.Task.sequence==sequence)
        count=members.count()
        self.release()
        return count
    
    def count_tasks_like(self, sequence, recovery, status):
        self.lock()
        with self.session.no_autoflush:
            members=self.session.query(self.Task).filter(self.Task.run_id==self.run_id, self.Task.sequence.like(sequence), self.Task.status.in_(status), self.Task.recovery==recovery)
        count=members.count()
        self.release()
        return count
    
    def get_task_status(self, task_names, sequence, recovery):
        self.lock()
        tasks=self.session.query(self.Task).filter(self.Task.run_id==self.run_id, self.Task.step_id.in_(task_names), self.Task.sequence==sequence, self.Task.recovery==recovery).all()
        result=dict()
        for task in tasks:
            result[task.step_id]=task.status        
        self.release()
        module_logger.debug("get_task_status: task: %s" %(repr(result)))
        return result
 
    def add_delay(self, delay_id, sequence, seconds, active=None, activated=None, recovery=None):
        delay=self.Delay(run_id=self.run_id, delay_id=delay_id, seconds=seconds, sequence=sequence, recovery=recovery,)
        module_logger.debug('DBAPI - add_delay: %s' % (repr(delay), ))
        self.lock()
        try:
            self.session.add(delay)
        except IntegrityError:
            self.release()
            raise
        self.commit_db()
        self.release()
        return delay_from_db(delay)
        
    def add_delay_update_if_not_exists(self, delay_id, sequence, seconds, active=None, activated=None, recovery=None):
        self.lock()
        delay=self.session.query(self.Delay).filter(self.Delay.run_id==self.run_id, self.Delay.sequence==sequence, self.Delay.delay_id == delay_id, self.Delay.recovery==recovery)
        found = self. session.query(delay.exists()).scalar()
        if not found:
            delay=self.Delay(run_id=self.run_id, delay_id=delay_id, seconds=seconds, sequence=sequence, recovery=recovery, active=active, activated=activated)
            self.session.add(delay)
            self.commit_db()
        elif delay.active != active:
            if active: delay.activated=datetime.utcnow()
            delay.active=active
            self.commit_db()
        self.release()
        return delay_from_db(delay)
        
    def get_delay_map(self, recovery=0):
        self.lock()
        items = self.session.query(self.Delay).filter(self.Delay.run_id==self.run_id, self.Delay.recovery==recovery).all()
        self.release()
        item_map=dict()
        for item in items:
            try:
                sequence_map=item_map[item.sequence]
            except KeyError:
                sequence_map=dict()
                item_map[item.sequence]=sequence_map
            sequence_map[item.delay_id]=delay_from_db(item)
        return item_map

    def get_delay_iter(self, recovery, active=True):
        self.lock()
        rows = self.session.query(self.Delay).filter(self.Delay.run_id==self.run_id, self.Delay.recovery==recovery).all()
        #rows = query.statement.execute().fetchall()
        self.release()
        for row in rows:
            yield delay_from_db(row)
  
    def _get_delay(self, delay):
        #self.lock()
        rows = self.session.query(self.Delay).filter(self.Delay.id_==delay.id_)
        #rows = query.statement.execute().fetchall()
        #self.release()
        try: return rows[0]
        except: return None
  
    def activate_delay(self, delay):
        self.lock()
        db_delay = self._get_delay(delay)
        delay.activated=db_delay.activated=datetime.utcnow()
        delay.active=True
        self.commit_db()
        self.release()
        return delay
        
    def deactivate_delay(self, delay):
        self.lock()
        db_delay = self._get_delay(delay)
        delay.active=db_delay.active=False
        self.commit_db()
        self.release()
        return delay

    def count_active_delays(self, sequence, recovery,):
        self.lock()
        with self.session.no_autoflush:
            members=self.session.query(self.Delay)
            members=members.filter(self.Delay.run_id==self.run_id, self.Delay.recovery==recovery, self.Delay.active.is_(True))
        
        now=datetime.utcnow()
        time_to_mature=[m.seconds-(now-m.activated).total_seconds() for m in members]
        min_time_to_mature = min(time_to_mature) if len(time_to_mature) > 0 else None
        count=members.count()
        self.release()
        return count, min_time_to_mature
    
        
 
if __name__ == '__main__':
    import pickle
    #file='/var/acrisel/sand/eventor/eventor/eventor/eventor/schema.db'
    #file=':memory:'
    mydb=DbApi(userstore='sqfile', root='eventor.databases', config='dbapi.conf')
    # mydb.add_event(event_id='34', name='evently')
    mydb.commit_db()
    
    delay=mydb.add_delay(delay_id='mydelay', seconds=2419200, sequence=3, )
    print(repr(delay))
    delay=mydb.activate_delay(delay)
    print(repr(delay))
    delay=mydb.deactivate_delay(delay)
    print(repr(delay))
    mydb.commit_db()
    task=mydb.Task()




# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
"""
 
About
=========
:synopsis:     commandline  handler for existing eventor db
:moduleauthor: Arnon Sela @ acrisel.com
:date:         Nov 26, 2016
:description:  cmd provides access to set of commands to manipulate eventor db.  
               This would be done after a run and before rerun.


Command Line Arguments:
----------------------

.. cmdoption::  <Fire Perimeter> 

      Fire perimeter from FIMT Schema (Feature Class)
    
.. cmdoption::  <Fire to Export> 

      List of fires to export (The fire number in the feature class above that you want to load)
      
Example to run on the command line::

    running_module.py //some/url some.other.argument 
    

Outputs:
-------------------
If you script creates any outputs delete this line here and replace it 
with a description of the outputs produced by your script.  If your script
does not produce any outputs delete the entire output section!

Dependencies:
-------------------
Does your script have any dependencies?  If it does then include them here.  Only
include dependencies like 3rd party modules that your script requires that are not 
part of its hosting project.


**History:**
-------------------

:Author: <who made the changes>
:Modification:
   - Describe here the changes that were made in bullet form.  Use
     a new dash character '-' to create a new entry!
:Date: <enter the date the changes were made>


API DOC:
===============     
"""
# TODO:
# handler main commands



'''
Created on Nov 23, 2016

@author: arnon
'''

from enum import Enum
from acris import threaded

class EventorError(Exception):
    pass

class AssocType(Enum):
    event=1
    step=2
    
    
class TaskStatus(Enum):
    ''' Eventor internal task states
    '''
    ready=1     # triggered to run by events, waiting for resources
    allocate=2  # resource allocation, waiting for allocation of resources
    fueled=3   # resource satisfied, waiting to be activated
    active=4    # running
    success=5   # finished successfully
    failure=6   # failed to finish
    
    
class StepStatus(Enum):
    ''' Programmer Step interface 
    '''
    ready=1
    allocate=2
    fueled=3
    active=4
    success=5
    failure=6
    complete=7 # finished successfully or with failure
    
def task_to_step_status(status):
    value=status.value
    result=StepStatus(value)
    return result

def step_to_task_status(status):  
    '''
    This assumes that complete status was already converted to failure and success
    '''
    value=status.value
    result=TaskStatus(value)
    return result
    
class StepReplay(Enum):
    rerun=1
    skip=2

class RunMode(Enum): 
    restart=1
    recover=2
    continue_=3 
    replay=4
    
class DbMode(Enum):
    write=1
    append=2
    read=3
    
class LoopControl(Enum):
    stop=1
    pause=2
    start=3
    resume=4
    kill=5  
    
class Invoke(object):
    '''
    Invoke is made to support looping super tasks.
    '''
    def __init__(self, target, args=(), kwargs={}):
        self.target=target
        self.args=args
        self.kwargs=kwargs
        self.is_alive_flag=False
        
    def start(self,):
        self.is_alive_flag=True
        result=self.target(*self.args, **self.kwargs)
        self.is_alive_flag=False
        return result
    
    @threaded
    def join(self, timeout=None):
        return
    
    @threaded
    def is_alive(self):
        return self.is_alive_flag


'''
Created on Oct 31, 2016

@author: arnon
'''

from eventor.eventor_types import EventorError
import logging
from _ast import arg

module_logger=logging.getLogger(__name__)

def expr_to_str(*args):
    items=list()
    for arg in args:
        if type(arg) == str:
            result=arg
        elif isinstance(arg, Event):
            result=arg.id_
        elif type(arg) == tuple:
            result= "(" + expr_to_str(*arg) +")" 
        else:
            raise EventorError("unknown variable in logical operation: %s" % repr(arg))
        items.append(result)
            
    expr="(" + ") and (".join(items) + ")"
    return expr
    

def or_(*args):
    items=list()
    for arg in args:
        result= expr_to_str(arg)
        items.append(result)
            
    expr=" or ".join(items)
    return expr

class Event(object):
    """Event object used to reuse in user program.
        
        Event objects are created by Eventor.add_event programming interface or by the mechanism itself
    
        Attributes:
            N/A
            
        Methods:
            db_write: write event to db file
    
    """

    def __init__(self, name, expr=None):
        self.name=name
        self.expr=expr
        if expr:
            self.expr=expr_to_str(expr)
        self.id_=name #get_event_id()
                    
    def __repr__(self):
        expr=''
        if self.expr:
            expr=", %s" % self.expr
        return "Event(%s%s)" % (self.id_, expr)
        
    def __str__(self):
        return repr(self)
    
    #def db_write(self, db):
    #    db.add_event(event_id=self.id_, name=self.name,)
    
    def trigger_(self, db, sequence):
        db.add_trigger(event_id=self.id_, sequence=sequence)
    
    def trigger_if_not_exists(self, db, sequence, recovery):
        added=db.add_trigger_if_not_exists(event_id=self.id_, sequence=sequence, recovery=recovery)
        return added
    
if __name__ == '__main__':
    e1=Event('E1')
    e2=Event('E2')
    e3=Event('E3')
    e4=Event('E4')
    
    expr=expr_to_str(or_(e1,e2),or_(e3,e4))
    print(expr, eval(expr, globals(), 
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True,}))
    
    expr=expr_to_str(e3,e4)
    print(expr, eval(expr, globals(), 
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True,}))
    
    expr=expr_to_str((e1,e2))
    print(expr, eval(expr, globals(), 
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True,}))
    


'''
Created on Oct 31, 2016

@author: arnon
'''

from .step import Step
import logging
#from .dbschema import Assoc as AssocDb
from .event import Event
from .eventor_types import EventorError

module_logger=logging.getLogger(__name__)


class Assoc(object):
    
    def __init__(self, event, assoc_obj):
        if not isinstance(event, Event):
            raise EventorError('event argument must be Event, but found %s' % type(event))
        if not isinstance(assoc_obj, (Event, Step)):
            raise EventorError('assoc_obj argument must be Event or Step, but found %s' % type(assoc_obj))
        
        self.event=event
        self.assoc_obj=assoc_obj
                    
    def __repr__(self):
        return "Assoc(%s, %s)" % (repr(self.event), repr(self.assoc_obj))
        
    def __str__(self):
        return "Assoc(%s, %s)" % (str(self.event), str(self.assoc_obj))

    #def db_write(self, db):
    #    db.add_assoc(event_id=self.event.id, obj_type=type(self.assoc_obj), obj_id=self.assoc_obj.id)


from .VERSION import __version__

from .main import Eventor
from .eventor_types import AssocType, TaskStatus, DbMode, StepStatus, StepReplay, RunMode, Invoke
from .event import or_
from .utils import calling_module, store_from_module


'''
Created on Jun 20, 2017

@author: arnon
'''

from sqlalchemy.engine.url import URL
from sqlalchemy import create_engine
from sqlalchemy.schema import MetaData 
from sqlalchemy.orm import sessionmaker, scoped_session
import logging
from eventor.conf_handler import getrootconf
import sys
import sqlite3

logger=logging.getLogger(__file__)

URL_ARGS=['drivername', 'username', 'password', 'host', 'port', 'database', 'query', 'file']
CONNECT_ARGS=['dialect', 'schema'] + URL_ARGS


def getdbinfo(conf=None, database=None, root='DATABASES', quiet=False, expand=True, expand_map=None):
    databases=getrootconf(conf=conf, root=root, expand=expand, expand_map=expand_map)
    db=None
    if database is not None:
        try: db=databases[database]
        except: 
            if not quiet:
                raise Exception("Database configuration not found for: %s"%(database))
            else: return None
    else: db=databases
    return db

def get_item(conf, tag, default=None, ):
    try:
        value=conf[tag]
    except Exception as e: 
        if not default:
            raise Exception("missing %s" % (tag)) from e
        value=default
    return value


def sqlite_url(dbconf, database):
    '''
    engine = create_engine('sqlite:///foo.db')
    engine = create_engine('sqlite:////absolute/path/to/foo.db')
    engine = create_engine('sqlite:///C:\\path\\to\\foo.db')
    engine = create_engine(r'sqlite:///C:\path\to\foo.db')
    engine = create_engine('sqlite://')
    '''
    creator=None
    
    file=dbconf.get('database')
    url_args=get_url_kwargs(dbconf, )
    url=URL(drivername='sqlite', **url_args)
    if not file:
        PY2 = sys.version_info.major == 2
        if PY2:
            params = {}
        else:
            params = {'uri': True}
            
        url_args_list=["%s=%s" % (k, v) for k, v in url_args.items()]
        url_args_str=''
        if url_args_list: url_args_str="?%s" % '&'.join(url_args_list)
        DB_URI = 'file::memory:%s' %(url_args_str)
        creator = lambda: sqlite3.connect(DB_URI, **params)
        
    kwargs=get_rest_kwargs(dbconf, ignore=CONNECT_ARGS, )
    if creator:
        kwargs['creator']=creator
    return url, kwargs

dialect_map={
    'sqlite': sqlite_url,
    }


def get_rest_kwargs(dbconf, ignore=[], ):
    kwargs=dict([(k,v) for k, v in dbconf.items() if k not in ignore])
    return kwargs

def get_url_kwargs(dbconf, ):
    kwargs=dict([(k,v) for k, v in dbconf.items() if k in URL_ARGS and k != 'drivername'])
    return kwargs

class SQLAlchemyConf(object):
    def __init__(self, conf=None, root='DATABASES', database=None, expand=True, expand_map=None, quiet=False, echo=False):
        self.conf=conf
        self.database=database
        self.dbconf=getdbinfo(conf=self.conf, root=root, database=self.database, expand=expand, expand_map=expand_map, quiet=quiet) 
        self.engine=None
        self.metadata=None
        self.session=None
        self.echo=echo
        
    def __repr__(self):
        return repr(self.dbconf)
        
    def get_info(self):
        return self.dbconf
          
    def get_url(self):
        dbconf=self.dbconf
        try:
            dialect=get_item(dbconf, 'dialect', )
        except Exception as e:
            raise Exception("dialect key missing in configuration for: %s" % (self.database)) from e

        func=dialect_map.get(dialect)
        
        if func:
            #except Exception as e:
            #raise Exception('unhandled dialect: %s, allow %s.' %(dialect, repr(list(dialect_map.keys())))) from e
            logger.debug('specialize map: %s' % repr(func))
            url, kwargs=func(dbconf, self.database)
        else:
            url_args=get_url_kwargs(dbconf, )
            try:
                url=URL(drivername=dialect, **url_args)
                # fetch other connection args (called query in sqlalchemy.engine.url.URL
            except Exception as e:
                raise Exception('Failed to create URL for : %s:%s' % (dialect, url_args.get('drivername'))) from e
            kwargs=get_rest_kwargs(dbconf, ignore=CONNECT_ARGS, )
            
        return url, kwargs
                    
    def get_engine(self, force=False):
        if self.engine and not force: return self.engine
        url, kwargs =self.get_url()
        self.engine=engine=create_engine(url, **kwargs, echo=self.echo, )
        return engine
        
    
    def get_metadata(self, force=False):
        if self.metadata and not force: return self.metadata
        dbconf=self.dbconf
        engine = self.get_engine(force=force)
        schema=dbconf.get('schema')
        self.metadata=metadata=MetaData(bind=engine, schema=schema)
        return metadata
    
    def get_session(self, force=False):
        if self.session and not force: return self.session
        session_factory=sessionmaker(bind=self.get_metadata(force=force).bind.engine)
        Session = scoped_session(session_factory)
        self.session=session=Session()
        return session

    
    

if __name__ == '__main__':
    url, connect_args=SQLAlchemyConf('db.conf', database='default').get_url()
    print(url, connect_args)
    
    dbconf=SQLAlchemyConf(conf='db.conf', database='playpg').get_info()
    url, connect_args=SQLAlchemyConf(dbconf,).get_url()
    print(url, connect_args)
    
    dbconf=SQLAlchemyConf(conf='db.conf', database='playmem').get_info()
    url, connect_args=SQLAlchemyConf(dbconf,).get_url()
    print(url, connect_args)

    dbconf=SQLAlchemyConf(conf='db.conf', database='playfile').get_info()
    url, connect_args=SQLAlchemyConf(dbconf,).get_url()
    print(url, connect_args)
    
    url, connect_args=SQLAlchemyConf('db.conf', database='playmy').get_url()
    print(url, connect_args)

    


'''
Created on Oct 19, 2016

@author: arnon
'''
import logging
import pprint

from .eventor_types import EventorError, TaskStatus
#from .utils import decorate_all, print_method_name

module_logger=logging.getLogger(__name__)

#class Step(metaclass=decorate_all(print_method_name)):
class Step(object):
    """A step in steps structure.  
       
        Each step has unique id that identifies it within Steps.
        It 
       
        A step is considered completed if it returns gradior.complete.
        step can also return gradior.failed or gradior.active.
       
        If step generate exception, it is considered gradior.FAILED.  The exception is
        also registered and could be referenced.   
    """

    def __init__(self, name=None, func=None, func_args=[], func_kwargs={}, triggers={}, acquires=None, releases=None, recovery={}, config={}):
        '''
        Constructor
        '''
        
        self.name=name
        self.id_=name #get_step_id()
        self.func=func
        self.func_args=func_args
        self.func_kwargs=func_kwargs
        self.triggers=triggers
        self.recovery=recovery
        self.config=config
        #self.pass_sequence=pass_sequence
        self.concurrent=0
        self.acquires=acquires
        self.releases=releases if releases is not None else acquires
        
        if func is not None and not callable(func):
            raise EventorError('func must be callable: %s' % repr(func))
        
        self.path=None
        self.iter_path=None
        
    def __repr__(self):
        if hasattr(self.func, '__name__'):
            fname=self.func.__name__
        else:
            fname=self.func.__class__.__name__
        
        #triggers=', '.join([pprint.pformat(t) for t in self.triggers])
        triggers=pprint.pformat(self.triggers)
        return "Step( name({}), func({}), triggers({}))".format(self.name, fname, triggers)
    
    def __str__(self):
        return repr(self)
    
    def _name(self, seq_path):
        result='/'
        if self.name:
            result="%s/%s" %(self.name, seq_path)
        return result
             
    def db_write(self, db):
        db.add_step(step_id=self.id_, name=self.name)
    
    def trigger_(self, db, sequence):
        db.add_task(event_id=self.id_, sequence=sequence,status=TaskStatus.ready)
    
    def trigger_if_not_exists(self, db, sequence, status, recovery=None):
        added=db.add_task_if_not_exists(step_id=self.id_, sequence=sequence, status=status, recovery=recovery)
        return added
    
    def __call__(self, seq_path=None, loop_value=None,):
        module_logger.debug('[ Step %s ] Starting: %s' % (self._name(seq_path), repr(self) ))
        if self.func is not None:
            func=self.func
            func_args=self.func_args
            func_kwargs=self.func_kwargs
            sequence_arg_name=self.config['sequence_arg_name']
            if sequence_arg_name:
                self.func_kwargs.update({sequence_arg_name: seq_path})
            #if self.config['pass_resources']:
            #    self.func_kwargs.update({'eventor_task_resoures': resources})
            #stop_on_exception=self.config['stop_on_exception']
            result=func(*func_args, **func_kwargs)
        else:
            result=True
        module_logger.debug('[ Step %s ] Completed: %s' % (self._name(seq_path), repr(self) ))
        return result



import os
import sys

from setuptools import setup
from distutils.sysconfig import get_python_lib

''' 
is the Python package in your project. It's the top-level folder containing the 
__init__.py module that should be in the same directory as your setup.py file
/-
  |- README.rst
  |- CHANGES.txt
  |- setup.py
  |- dogs 
     |- __init__.py
     |- catcher.py

To create package and upload:

  python setup.py register
  python setup.py sdist
  twine upload -s dist/path/to/gz

'''
PACKAGE = "eventor"

'''
NAME is usually similar to or the same as your PACKAGE name but can be whatever you want. 
The NAME is what people will refer to your software as, the name under which your 
software is listed in PyPI and—more importantly—under which users will install it 
(for example, pip install NAME).
'''
NAME = PACKAGE

'''
DESCRIPTION is just a short description of your project. A sentence will suffice.
'''
DESCRIPTION = '''eventor is a python programming facility to program event based sequence of activities'''

'''
AUTHOR and AUTHOR_EMAIL are what they sound like: your name and email address. This 
information is optional, but it's good practice to supply an email address if people 
want to reach you about the project.
'''
AUTHOR = 'Acrisel Team'
AUTHOR_EMAIL = 'support@acrisel.com'

'''
URL is the URL for the project. This URL may be a project website, the Github repository, 
or whatever URL you want. Again, this information is optional.
'''
URL = 'https://github.com/Acrisel/eventor'

version_file=os.path.join(PACKAGE, 'VERSION.py')
with open(version_file, 'r') as vf:
    vline=vf.read()
VERSION = vline.strip().partition('=')[2].replace("'", "")
#VERSION =__import__(version_file).__version__

# Warn if we are installing over top of an existing installation. This can
# cause issues where files that were deleted from a more recent Accord are
# still present in site-packages. See #18115.
overlay_warning = False
if "install" in sys.argv:
    lib_paths = [get_python_lib()]
    if lib_paths[0].startswith("/usr/lib/"):
        # We have to try also with an explicit prefix of /usr/local in order to
        # catch Debian's custom user site-packages directory.
        lib_paths.append(get_python_lib(prefix="/usr/local"))
        
    for lib_path in lib_paths:
        existing_path = os.path.abspath(os.path.join(lib_path, PACKAGE))
        if os.path.exists(existing_path):
            # We note the need for the warning here, but present it after the
            # command is run, so it's more likely to be seen.
            overlay_warning = True
            break

setup_info={'name': NAME,
 'version': VERSION,
 'url': URL,
 'author': AUTHOR,
 'author_email': AUTHOR_EMAIL,
 'description': DESCRIPTION,
 'long_description': open("README.rst", "r").read(),
 'license': 'MIT',
 'keywords': 'project, virtualenv, parameters',
 'packages': [PACKAGE],
 'install_requires': [
     'SQLAlchemy>=1.1.5',
     'acris>=2.0.2',
     'acrilog',
     'setproctitle>=1.1',
     'pyyaml>=3.12',
     'namedlist>=1.7',
     'PyYAML>=3.12',
     ],
 'extras_require': {'dev': [], 'test': []},
 'classifiers': ['Development Status :: 5 - Production/Stable',
                 'Environment :: Other Environment',
                 #'Framework :: Project Settings and Operation',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.2',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Software Development :: Libraries :: Application Frameworks',
                 ]}
setup(**setup_info)


if overlay_warning:
    sys.stderr.write("""

========
WARNING!
========

You have just installed ProjEnv over top of an existing
installation, without removing it first. Because of this,
your install may now include extraneous files from a
previous version that have since been removed from
Accord. This is known to cause a variety of problems. You
should manually remove the

%(existing_path)s

directory and re-install ProjEnv.

""" % {"existing_path": existing_path})



'''
Created on Oct 19, 2016

@author: arnon
'''

from acris import Sequence
from logging.handlers import QueueListener
import inspect
import datetime
import os
import yaml


def is_require_op(op):
    if op in ['or', 'and',]:
        return True
    return False

def op_to_lambda(op):
    return "lambda x,y: x {} y".format(op)

StepId=Sequence('_EventorStepId')
EventId=Sequence('_EventorEventId')
DelayId=Sequence('_EventorDelayId')

def rest_sequences():
    StepId.reset()
    EventId.reset()
    DelayId.reset()

def get_step_id():
    return "S%s" % StepId()

def get_event_id():
    return "E%s" % EventId()


def get_delay_id():
    return "d%s" % DelayId()


def valid_step_name(name):
    return name.find('.') == -1


class CustomQueueListener(QueueListener):
    def __init__(self, queue, *handlers):
        super(CustomQueueListener, self).__init__(queue, *handlers)
        """
        Initialise an instance with the specified queue and
        handlers.
        """
        # Changing this to a list from tuple in the parent class
        self.handlers = list(handlers)
        

    def handle(self, record):
        """
        Override handle a record.

        This just loops through the handlers offering them the record
        to handle.

        :param record: The record to handle.
        """
        record = self.prepare(record)
        for handler in self.handlers:
            if record.levelno >= handler.level: # This check is not in the parent class
                handler.handle(record)

    def addHandler(self, hdlr):
        """
        Add the specified handler to this logger.
        """
        if not (hdlr in self.handlers):
            self.handlers.append(hdlr)
            

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this logger.
        """
        if hdlr in self.handlers:
            hdlr.close()
            self.handlers.remove(hdlr)

        
def traces(trace):
    '''
     File "/private/var/acrisel/sand/gradior/gradior/gradior/gradior/loop_task.py", line 41, in task_wrapper
    '''
    result=[ "File \"%s\", line %s, in %s\n    %s" % (frame.filename, frame.lineno, frame.function, frame.code_context[0].rstrip()) for frame in trace]
    return result


def calling_module(depth=2):
    frame_records = inspect.stack()[depth]
    return frame_records.filename

def store_from_module(module, module_location=False):
    location=os.path.dirname(module)
    name=os.path.basename(module)
    
    parts=name.rpartition('.')
    if parts[0]:
        if parts[2] == 'py':
            module_runner_file=parts[0]
        else:
            module_runner_file=name
    else:
        module_runner_file=parts[2]
    filename='.'.join([module_runner_file, 'run.db'])  
    
    if module_location:
        filename=os.path.join(location, filename)
    else:
        filename=os.path.join(os.getcwd(), filename)  
    return filename



from types import FunctionType

# check if an object should be decorated
def do_decorate(attr, value):
    # result = ('__' not in attr and
    result = (isinstance(value, FunctionType) and
              getattr(value, 'decorate', True))
    return result

# decorate all instance methods (unless excluded) with the same decorator
def decorate_all(decorator):
    class DecorateAll(type):
        def __new__(cls, name, bases, dct):
            for attr, value in dct.items():
                if do_decorate(attr, value):
                    decorated=decorator(name, value)
                    dct[attr] = decorated
                    #print('decorated', attr, decorated)
            return super(DecorateAll, cls).__new__(cls, name, bases, dct)
    return DecorateAll

# decorator to exclude methods
def dont_decorate(f):
    f.decorate = False
    return f

def print_method(print_func):
    def print_method_name(name, f):
        def wrapper(*args, **kwargs):
            print_func('entering method: %s.%s' % (name, f.__name__, ))
            start=datetime.datetime.now()
            result=f(*args, **kwargs)
            finish=datetime.datetime.now()
            print_func('exiting method: %s.%s, time span: %s' % (name, f.__name__, str(finish-start)))
            return result
        return wrapper
    return print_method_name

if __name__ == '__main__':

    meta_decorator=decorate_all(print_method(print))
    
    class Foo(metaclass=meta_decorator):

        def __init__(self):
            pass
        
        def bar(self):
            pass
        
        @dont_decorate
        def baz(self):
            pass
        
        @classmethod
        def test(self):
            pass
     
    foo=Foo()
    foo.bar()   
    foo.bar()   
    



'''
Created on Nov 23, 2016

@author: arnon
'''

import logging
from acris import Sequence, MergedChainedDict
from acrilog import MpLogger
from acris import virtual_resource_pool as vrp
import multiprocessing as mp
from collections import namedtuple
import inspect
from enum import Enum
import os
import pickle
import queue
import time
from collections import Mapping
from datetime import datetime
from eventor.step import Step
from eventor.event import Event
from eventor.delay import Delay
from eventor.assoc import Assoc
from eventor.dbapi import DbApi
from eventor.utils import calling_module, traces, rest_sequences, store_from_module, get_delay_id
from eventor.eventor_types import Invoke, EventorError, TaskStatus, step_to_task_status, task_to_step_status, LoopControl, StepStatus, StepReplay, RunMode, DbMode
from eventor.VERSION import __version__
from eventor.dbschema import Task
from eventor.conf_handler import getrootconf

try: 
    from setproctitle import setproctitle
except: 
    setproctitle=None
#from eventor.loop_event import LoopEvent

#module_logger=logging.getLogger(__name__)
module_logger=None

from acris import traced_method
from eventor.utils import decorate_all, print_method

#traced=traced_method(None, True)

        
import socket
def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    address = s.getsockname()[0]
    s.close()
    return address


def get_unique_run_id():
    ip = get_ip_address()
    ips = ''.join(ip.split('.'))
    now = datetime.now().strftime("%Y%m%d%H%M%S")
    result = "%s@%s@%s" % (os.getpid(), ips, now)
    return result

class TaskAdminMsgType(Enum):
    result=1
    update=2
    
    
TaskAdminMsg=namedtuple('TaskAdminMsg', ['msg_type', 'value', ])
TriggerRequest=namedtuple('TriggerRequest', ['type', 'value'])


class ResourceAllocationCallback(object):
    def __init__(self, notify_queue):
        self.q=notify_queue
        
    def __call__(self,resources=None):
        self.q.put(resources)


def task_wrapper(run_id=None, task=None, step=None, adminq=None, use_process=True, logger_info=None):
    ''' 
    Args:
        func: object with action method with the following signature:
            action(self, action, unit, group, sequencer)    
        action: object with taskid, unit, group: id of the unit to pass
        sqid: sequencer id to pass to action '''
    global module_logger
    
    if use_process:
        module_logger=MpLogger.get_logger(logger_info=logger_info, name='') # name="%s.%s_%s" %(logger_info['name'], step.name, task.sequence))
    
    task.pid=os.getpid()
    os.environ['EVENTOR_STEP_SEQUENCE']=str(task.sequence)
    os.environ['EVENTOR_STEP_RECOVERY']=str(task.recovery)
    os.environ['EVENTOR_STEP_NAME']=str(step.name)

    if setproctitle is not None and use_process:
        run_id_s = "%s." % run_id if run_id else ''
        setproctitle("eventor: %s%s.%s(%s)" % (run_id_s, step.name, task.id_, task.sequence))

    # Update task with PID
    update=TaskAdminMsg(msg_type=TaskAdminMsgType.update, value=task) 
    adminq.put( update )
    module_logger.info('[ Step {}/{} ] Trying to run'.format(step.name, task.sequence))
    
    try:
        # todo: need to pass task resources.
        result=step(seq_path=task.sequence, )
    except Exception as e:
        trace=inspect.trace()
        trace=traces(trace) #[2:]
        task.result=(e, pickle.dumps(trace))
        task.status=TaskStatus.failure
    else:
        task.result=result
        task.status=TaskStatus.success
        
    result=TaskAdminMsg(msg_type=TaskAdminMsgType.result, value=task) 
    module_logger.info('[ Step {}/{} ] Completed, status: {}'.format(step.name, task.sequence, str(task.status), ))
    adminq.put( result )
    return True


class EventorState(Enum):
    active=1
    shutdown=2
    
    
class RpCallback(object):
    def __init__(self, notify_queue, task_id=''):
        self.q=notify_queue
        self.task_id=task_id
    #def get_name(self):
    #    return self.name
    def __call__(self,ready=False):
        self.q.put( self.task_id )


class Memtask(object):
    def __init__(self, task, ):
        self.id_=task.id_
        self.step_id=task.step_id
        self.sequence=task.sequence
        self.requestor=None
        self.fueled=False
        self.request_id=None
        #self.resources=None
        
    #def __del__(self):
    #    if self.requestor and self.resources:
    #        self.requestor.put(*self.resources)
    
#class Eventor(metaclass=decorate_all(print_method(module_logger.debug))):
class Eventor(object):
    """Eventor manages events and steps that needs to be taken when raised.
    
    Eventor provides programming interface to create events, steps and associations among them.
    It provides means to raise events, and a service that would perform steps according to those events. 
        
    Attributes: 
        controlq: mp queue is used to control endless eventor loop
    
    Methods:
        add_event: 
        add_step: 
    
    """
    
    config_defaults={'workdir':'/tmp', 
                     'logdir': '/tmp', 
                     'task_construct': mp.Process, 
                     #'synchrous_run': False, 
                     'max_concurrent': -1, 
                     'stop_on_exception': True,
                     'sleep_between_loops': 0.1,
                     'sequence_arg_name': None, # 'eventor_task_sequence'
                     #'pass_resources': False,
                     }  
    
    recovery_defaults={StepStatus.ready: StepReplay.rerun, 
                      StepStatus.allocate: StepReplay.rerun, 
                       StepStatus.fueled: StepReplay.rerun, 
                       StepStatus.active: StepReplay.rerun, 
                       StepStatus.failure: StepReplay.rerun, 
                       StepStatus.success: StepReplay.skip,}  
        
    def __init__(self, name='', store='', run_mode=RunMode.restart, recovery_run=None, dedicated_logging=False, logging_level=logging.INFO, run_id='', shared_db=False, config={},):
        """initializes steps object
        
        Args:
            name (str): human readable identifying Eventor among other eventors
            store (path): file in which Eventor data would be stored and managed for reply/restart 
                if the value is blank, filename will be based on calling module 
                if the value is :memory:, an in-memory temporary structures will be used

            run_mode: (RunMode) set Eventor to operate in recovery, restart, or continue
            recovery_run (int): since 'store' maintain historical runs, recovery can engage any
                previous run.  recovery_run, if provided, will tell Eventor which run to recover.
                If not provided, latest run is assumed.
                
            shared_db: (boolean) if set, indicates that the database used is by multiple 
                programs or instances thereof. 
                
            run_id: (str) if shared_db, run_id must be unique program execution among all programs 
                    sharing.  If not provided and shared_db is set, unique run_id will be generated.
                    When in recovery, and shared_db is set, run_id must be provided to identify the
                    program and its run to be recovered.

            config: configuration parameters to be used in operating eventor.
            can be file with YAML style configuration or a dictionary
                parameters can include the following keys:
                - logdir=/tmp, 
                - workdir=/tmp, 
                #- synchrous_run=False,
                - task_construct=mp.Process, 
                - stop_on_exception=True,
                - sleep_between_loops=1
                - databases: if config is provided, 
                    it will try to fetch store or name as code for database.
                    if both not provided, will try default database code.
                    if not found, will treat store and name as above.
                                                                  
                Example:
                    DATABASES:
                        default: 
                            dialect:  postgresql
                            drivername :  psycopg2
                            username: mydatabaseuser
                            password: mydatabasepass
                            host:     localhost
                            port:     5433
                            database: postgres
                            schema:   eventor
                    
                        playfile:
                            dialect: sqlite
                            file: /var/database/alchemy/schema.db  


        Returns:
            new object
            
        Raises:
            N/A
        """
        global module_logger
        
        self.name=''
        config_root_name = os.environ.get('EVENTOR_CONFIG_TAG', 'EVENTOR')
        if isinstance(config, str):
            frame = inspect.stack()[1]
            module = inspect.getsourcefile(frame[0])
            config_path = os.path.join(os.path.dirname(module), config)
            if os.path.isfile(config_path):
                config=config
            
        rootconfig = getrootconf(conf=config, root=config_root_name)
        self.__config = MergedChainedDict(rootconfig, os.environ, Eventor.config_defaults) 
        self.__controlq = mp.Queue()
        self.__steps=  dict() 
        #self.__acquires=dict() 
        #self.__releases=dict() 
        self.__tasks = dict() 
        self.__events = dict() 
        self.__assocs = dict()
        self.__delays = dict()
        self.__rp_notify = queue.Queue()
        
        #if dedicated_logging:
        #    logging_root = '.'.join(__name__.split('.')[:-1])
        #else:
        #    logging_root = ''
        level_formats = {logging.DEBUG:"[ %(asctime)-15s ][ %(processName)-11s ][ %(levelname)-7s ][ %(message)s ][ %(module)s.%(funcName)s(%(lineno)d) ]",
                        'default':   "[ %(asctime)-15s ][ %(processName)-11s ][ %(levelname)-7s ][ %(message)s ]",
                        }
        
        # TODO(Arnon): get to logging configuration (as in Database)
        # TODO(Arnon): drive encoding from parameter
        logger_name = ''
        if dedicated_logging:
            logger_name = __name__
        self.__logger = MpLogger(name=__name__, logging_level=logging_level, level_formats=level_formats, datefmt='%Y-%m-%d,%H:%M:%S.%f', logdir=self.__config['logdir'], encoding='utf8')
        module_logger = self.__logger.start()
        
        self.__calling_module = calling_module()
        self.store = store
        self.debug = logging_level == logging.DEBUG

        self.__adminq_mp_manager = mp_manager=mp.Manager()
        self.__adminq_mp = mp_manager.Queue()
        self.__adminq_th = queue.Queue()
        #self.__resource_notification_queue=mp.Queue()
        self.__requestq = queue.Queue()
        self.__task_proc = dict()
        self.__state = EventorState.active
        self.__run_mode = run_mode
        self.__recovery_run = recovery_run
        self._session_cycle_loop = False
        
        self.shared_db = shared_db
        self.run_id = run_id if run_id is not None else ''
        if shared_db and not self.run_id:
            if run_mode != RunMode.restart:
                raise EventorError("When shared_db is set in restart, run_id must be provided.")
            # in this case we need to produce unique run_id on this cluster
            self.run_id = get_unique_run_id()
            module_logger.info("Created process run_id: %s" % self.run_id) 
        rest_sequences()   
        self.__setup()
                
    def __setup(self):
        global module_logger
        filename = store_from_module(self.__calling_module)
        #module_logger.info("Eventor store file: %s" % filename)
        
        db_mode = DbMode.write if self.__run_mode==RunMode.restart else DbMode.append
        #self.__db_daly_adj=0
        #if self.__run_mode != RunMode.restart:
        #    try:
        #        db_mtime=os.path.getmtime(self.__filename)
        #    except OSError:
        #        pass
        #    else:
        #        self.__db_daly_adj=(datetime.now() - datetime.fromtimestamp(db_mtime)).total_seconds()
        self.db = DbApi(config=self.__config, modulefile=filename, shared_db=self.shared_db, run_id=self.run_id, userstore=self.store, mode=db_mode, echo=False, logger=module_logger) #self.debug)
        self.__requestors = vrp.Requestors()
        if self.__run_mode == RunMode.restart:
            self.__write_info()
        else:
            self.__read_info(run_mode=self.__run_mode, recovery_run=self.__recovery_run)
        
        
    def __repr__(self):
        steps='\n'.join([ repr(step) for step in self.__steps.values()])
        events='\n'.join([ repr(event) for event in self.__events.values()])
        assocs='\n'.join([ repr(assoc) for assoc in self.__assocs.values()])
        result="Steps( name( {} ) events( {} ) steps( {} ) assocs( {} )  )".format(self.name, events, steps, assocs)
        return result
        
    def __str__(self):
        steps='\n    '.join([ str(step) for step in self.__steps.values()])
        events='\n    '.join([ str(event) for event in self.__events.values()])
        assocs='\n    '.join([ str(assoc) for assoc in self.__assocs.values()])
        result="Steps( name( {} )\n    events( \n    {}\n   )\n    steps( \n    {}\n   )\n    assocs( \n    {}\n   )  )".format(self.name, events, steps, assocs)
        return result

    def _name(self, seq_path):
        result='/'
        if self.name:
            result="%s/%s" %(self.name, seq_path)
        return result
        
    def __write_info(self, run_file=None):
        self.__recovery=0
        info={'version': __version__,
              'program': self.__calling_module,
              'recovery': self.__recovery,
              }
        self.db.write_info(**info)
        #self.previous_triggers=None
        self.__previous_tasks=None
        self.__previous_delays=None
    
    def __read_info(self, run_mode=None, recovery_run=None):
        self.__info=self.db.read_info()
        recovery=recovery_run
        if recovery is None:
            previous_recovery=self.__info['recovery']
        if self.__run_mode == RunMode.recover:
            recovery=str(int(previous_recovery) + 1)
            self.db.update_info(recovery=recovery)
        else:
            recovery=previous_recovery
        self.__recovery=recovery
        #self.previous_triggers=self.db.get_trigger_map(recovery=recovery)
        self.__previous_tasks=self.db.get_task_map(recovery=previous_recovery)
        self.__previous_delays=self.db.get_delay_map(recovery=previous_recovery)
            
    def __convert_trigger_at_complete(self, triggers):
        at_compete=triggers.get(StepStatus.complete)
        if at_compete:
            del triggers[StepStatus.complete]
            at_fail=triggers.get(StepStatus.failure, list())
            at_success=triggers.get(StepStatus.success, list())
            at_fail.extend(at_compete)
            at_success.extend(at_compete)
            triggers[StepStatus.failure]=at_fail
            triggers[StepStatus.success]=at_success
        return triggers
          
    def __convert_recovery_at_complete(self, recovery):
        at_compete=recovery.get(StepStatus.complete)
        if at_compete is not None:
            del recovery[StepStatus.complete]
            at_fail=recovery.get(StepStatus.failure, at_compete)
            at_success=recovery.get(StepStatus.success, at_compete)
            recovery[StepStatus.failure]=at_fail
            recovery[StepStatus.success]=at_success
        return recovery

    def add_event(self, name, expr=None):
        """add a event to Eventor object
        
        Args:
            name: (string) unique identifier provided by caller
            
        returns:
            new event that was added to Eventor; this event can be used further in assoc method
        """
        try:
            event=self.__events[name]
        except:
            pass
        else:
            if expr == event.expr:
                return event
            
        event=Event(name, expr=expr)
        self.__events[event.id_]=event
        module_logger.debug('add_event: %s' %( repr(event), ))
        return event
    
    def get_step(self, name):
        return self.__steps.get(name, None)
    
    def add_step(self, name, func=None, args=(), kwargs={}, triggers={}, acquires=None, releases=None, recovery={}, config={}):
        """add a step to steps object
    
        config parameters can include the following keys:
            - stop_on_exception=True
            - sequence_arg_name=None : when set, corresponding name will be used as keyword argument to passed sequence to step
            #- pass_resources=False: when set, eventor_task_resources keyword argument will be passed to step
            
        Args:
            name: (string) unique identifier
            func: (callable) if provided, will be called when step is activated; otherwise its nope step.
            func_args: args to pass step when executing
            func_kwargs: keyword args to pass step when executing
            config: additional dict of keywords configuration 
            triggers: set of events to trigger once step processing is done
            acquires: list of resources to acquire before starting
            releases: list of resources to release once done; 
                defaults to acquires. if not provided.
            recovery: instructions on how to deal with 
                default: {TaskStatus.failure: StepReplay.rerun, 
                          TaskStatus.success: StepReplay.skip}
            
        returns:
            new step that was added to Eventor; this step can be used further in assoc method
            
        raises:
            EventorError: if func is not callable
        """
        try:
            step=self.__steps[name]
        except:
            pass
        else:
            return step
        
        triggers=self.__convert_trigger_at_complete(triggers)
        recovery=self.__convert_recovery_at_complete(recovery)
        recovery=MergedChainedDict(recovery, Eventor.recovery_defaults)
        recovery=dict([(step_to_task_status(status), replay) for status, replay in recovery.items()])
        
        config=MergedChainedDict(config, self.__config, os.environ,)
        step=Step(name=name, func=func, func_args=args, func_kwargs=kwargs, acquires=acquires, releases=releases, config=config, triggers=triggers, recovery=recovery)
        found=self.__steps.get(step.id_)
        if found:
            raise EventorError("Step with similar name already defined: %s" % step.id_)
        self.__steps[step.id_]=step
        module_logger.debug('add_step: %s' %( repr(step), ))
        return step
    

    def __get_start_delay_task(self, delay_id, seconds,):
        ''' checks Delay table `
        '''
        def delay_func(sequence, recovery, activated=None, active=True):
            ''' Inserts delay into Delay table to be picked up by delay_loop
            '''
            #module_logger.debug('add_delay_if_not_exists: %s' %( repr(active), ))
            delay=self.db.add_delay_update_if_not_exists(delay_id=delay_id, sequence=sequence, seconds=seconds, recovery=recovery, active=active, activated=activated)
            module_logger.debug('add_delay_if_not_exists: %s' %( repr(delay), ))
            return True
        return delay_func
        
    def add_assoc(self, event, *assocs, delay=0):
        """add a assoc to Eventor object
        
        Associates event with one or more objects of steps and events.
        
        When delay is provided, Eventor will activated assocs only after delay 
        seconds passed after event was triggered.
        
        Delay supports ':memory:' store only if Eventor is called with negative max_loops (default).
        This otherwise, the information pertaining the delay may not be maintained. 
        
        Args:
            event: an event object return from add_event()
            assocs: list of either step or event objects returned from add_step() or add_event() respectively
            delay: seconds to wait before activating assocs once event had been triggered.
            
        returns:
            N/A
            
        raises:
            EnventorError: if event is not of event type or obj is not instance of Event or Step
        """
        
        if delay > 0:
            # since we have to delay, we need to create a Delay hidden task in 
            # between event and assocs.
            #delay_seq=Sequence('__delay_assoc')  
            delay_id="_evr_delay_%s_%s" % (event.name, get_delay_id())
            delay_event=self.add_event(delay_id)
            delay_func=self.__get_start_delay_task(delay_id, seconds=delay,)
            delay_step=self.add_step(delay_id, func=delay_func,) 
            self.add_assoc(event, delay_step)
            self.add_assoc(delay_event, *assocs)
            module_logger.debug("adding delayed assoc: %s(%s) and %s(%s)" % (event, delay_step, delay_event, repr(assocs)))
            self.__delays[delay_id]=Delay(delay_id=delay_id, func=delay_func, seconds=delay, event=delay_event)
            
        else:    
            try:
                objs=self.__assocs[event.id_]
            except KeyError:
                objs=list()
                self.__assocs[event.id_]=objs
    
            for obj in assocs:
                assoc=Assoc(event, obj)
                objs.append(assoc)
                module_logger.debug('add_assoc: %s' %( repr(assoc), ))
    
    def trigger_event(self, event, sequence=0, db=None):
        """Activates event 
        
            Activate event by registering it in trigger table. 
            
            If Sequence is provided, it will be used as trigger sequence for this event.
            This is done to allow multiple triggers for the same event.
            
            If sequence is not provided, one will assigned to this trigger.  
            
            Sequence creates association with derived steps or events associated with the event triggered.
        
            Args:
                event: (Event) object returned from add_event()
                sequence: (str) object uniquely identifying this trigger among other triggers of the same event
                
            Returns:
                N/A
            
            Raises:
                EventorError
                
        """
        
        if not db:
            db=self.db
        added=event.trigger_if_not_exists(db=db, sequence=sequence, recovery=self.__recovery)
        return added
        
    def remote_trigger_event(self, event, sequence=0,):
        trigger_request=TriggerRequest(type='event', value=(event, sequence))
        self.__requestq.put(trigger_request)
        
    def trigger_step(self, step, sequence):
        """Activates step 
        
            Activate step by registering it in task table are 'ready'.  Task loop will pick it up for processing. 
        
            Args:
                step: (Step) object returned from add_step()
                sequence: (str) object uniquely identifying this trigger among other triggers of the same event
                
            Returns:
                N/A
            
            Raises:
                EventorError
                
        """
        task=step.trigger_if_not_exists(self.db, sequence, status=TaskStatus.ready, recovery=self.__recovery)
        if task:
            self.__triggers_at_task_change(task)
        return task is not None
    
    def __loop_trigger_request(self):
        while True:   
            try:
                request=self.__requestq.get_nowait() 
            except queue.Empty:
                return 
            if request.type=='event':
                event, sequence = request.value
                self.trigger_event(event, sequence)
                module_logger.debug('[ Event %s/%s ] Triggering event' % (event.id_, sequence))
        
          
    def __assoc_loop(self, event, sequence):
        ''' Fetches event associations and trigger them.
        
        If association is an event:
            trigger the event
            
        If association is a step:  
            register task
            trigger event - task running.
            
        '''
        assoc_events=list()
        assoc_steps=list()
        try:
            assocs=self.__assocs[event.id_]
        except KeyError:
            assocs=[]
            
        for assoc in assocs:
            assoc_obj=assoc.assoc_obj
            if isinstance(assoc_obj, Event):
                # trigger event
                module_logger.debug('Processing event association event [%s]: %s' % (sequence, repr(assoc_obj)))
                self.trigger_event( assoc_obj, sequence )
                assoc_events.append(sequence)
            elif isinstance(assoc_obj, Step):
                # Check if there is previous task for this step
                module_logger.debug('Processing event association step [%s]: %s' % (sequence, repr(assoc_obj)))
                self.trigger_step(assoc_obj, sequence)
                assoc_steps.append(sequence)
                
            else:
                raise EventorError("Unknown assoc object in association: %s" % repr(assoc))
        
        return list(set(assoc_events)), list(set(assoc_steps))

    def __loop_event(self):
        loop_seq=Sequence('EventLoop')
        self.loop_id=loop_seq() 
        module_logger.debug("Going to fetch events: %s: recovery: %s" % (self.name, self.__recovery, ))    
        # first pick up requests and move to act
        # this step is needed so automated requests will not impact 
        # the process as it is processing.
        # requests not picked up in current loop, will be picked by the next.
        triggers=self.db.get_trigger_iter(recovery=self.__recovery)
        trigger_db=dict()
        
        event_seqs=list()
        step_seqs=list()

        # need to rearrange per iteration
        for trigger in triggers:
            try:
                trigger_map=trigger_db[trigger.sequence]
            except KeyError:
                trigger_map=dict()
                trigger_db[trigger.sequence] = trigger_map
            
            #print('trigger %s[%s]' % (trigger.event_id, trigger.iteration) )  
            trigger_map[trigger.event_id] = True

            if not trigger.acted:
                assoc_events, assoc_steps = self.__assoc_loop(self.__events[trigger.event_id], trigger.sequence)
                event_seqs.extend(assoc_events)
                step_seqs.extend(assoc_steps)
                self.db.acted_trigger(trigger)
                                  
        # trigger_map=dict([(trigger.event_id, True) for trigger in triggers])
        for sequence, trigger_map in trigger_db.items():
            #print('trigger_map (%s)' % iteration, trigger_map)
            
            for event in self.__events.values():
                if not event.expr: continue
                
                try:
                    result=eval(event.expr, globals(), trigger_map)
                except:
                    result=False
                    
                module_logger.debug("[ Event %s/%s ] Eval expr: %s = %s\n    %s" % (event.id_, sequence,event.expr, result, trigger_map))
                
                # update trigger map with result
                #trigger_map[event.event_id]=result
                
                if result==True: # and self.act:
                    # TODO: do we need to raise event.
                    added=self.trigger_event(event, sequence,)   
                    #print('loop_event_iteration', event.event_id, iteration, added)  
                    if added:  
                        module_logger.debug('[ Event %s/%s] Triggered event (%s ):\n    {}' % (event.id_, sequence, repr(event))) 
                        event_seqs.append(sequence)
                        
        return list(set(event_seqs)), list(set(step_seqs))                  
    
    def __log_error(self, task, stop_on_exception):
        logutil=module_logger.error if stop_on_exception else module_logger.warning
        err_exception, pickle_trace=task.result
        err_trace=pickle.loads(pickle_trace)
        
        logutil('Exception in run_action: \n    {}'.format(task,)) #)
        logutil("%s" % (repr(err_exception), ))
        trace='\n'.join([line.rstrip() for line in err_trace])
        if trace: logutil("%s" % (trace, ) )

    def __apply_task_result(self, task):
        #step=self.__steps[task.step_id]
        module_logger.debug('[ Task %s/%s ] applying task update to db\n    %s' % (task.step_id, task.sequence, repr(task), ))
        self.db.update_task(task=task)
        if task.status in [TaskStatus.success, TaskStatus.failure]:
            self.__release_task_resources(task)
        triggered=self.__triggers_at_task_change(task)
        return triggered
    
    def __play_result(self, act_result,):
        module_logger.debug('Result collected: \n    %s' % ( repr(act_result)) )  
        stop_on_exception=self.__config['stop_on_exception']
        
        result=True
        istask=isinstance(act_result.value, Task)
        module_logger.debug('Received %s(%s) to play (istask: %s)' %(type(act_result.value).__name__, repr(act_result.value), istask))
        if istask: 
            task=act_result.value   
            if act_result.msg_type==TaskAdminMsgType.result:
                delay_task=task.step_id.startswith('_evr_delay_')
                if not delay_task:
                    proc=self.__task_proc[task.id_]
                    module_logger.debug('[ Task %s/%s ] applying result, process: %s, is_allive: %s' % \
                                        (task.step_id, task.sequence, repr(proc), proc.is_alive()))
                    #if proc.is_alive():
                    if not isinstance(proc, Invoke):
                        module_logger.debug('[ Task %s/%s ] joining (exitcode=%s)' % (task.step_id, task.sequence, proc.exitcode))
                        if proc.is_alive():
                            proc.join()
                        #while proc.is_alive():
                        #    proc.join(0.05)
                        module_logger.debug('[ Task %s/%s ] joined' % (task.step_id, task.sequence, ))
                    del self.__task_proc[task.id_]  
                    module_logger.debug('[ Task %s/%s ] deleted' % (task.step_id, task.sequence, ))
                step=self.__steps[task.step_id]
                step.concurrent-=1
                triggered=self.__apply_task_result(task)
                module_logger.debug('[ Task %s/%s ] triggered: %s, stop_on_exception: %s, task.status: %s' % \
                                    (task.step_id, task.sequence, repr(triggered), repr(stop_on_exception), repr(task.status)))
                shutdown=(len(triggered) == 0 or stop_on_exception) and task.status == TaskStatus.failure 
                module_logger.debug('[ Task %s/%s ] shutdown: %s' % (task.step_id, task.sequence, shutdown))
                if task.status==TaskStatus.failure:
                    self.__log_error(task, shutdown)
                    if len(triggered) == 0:
                        self.__state=EventorState.shutdown
                        
                #if shutdown:
                #    module_logger.info("Stopping running processes") 
                    
                    result=False
            elif act_result.msg_type==TaskAdminMsgType.update: 
                self.db.update_task(task=task)
                # TODO: stop running processes            
        else:
            # TODO: need to deal with action
            module_logger.debug('skip play result; act_result value type not matched: %s' % ( repr(type(act_result.value))) )  
        
        return result
    
    def __collect_results(self,):
        ''' Collect results from task wrapper queues.
        
        There are two queues for the different types of process constructs.
        multiprocess.Process: adminq_mp
        threading.Thread and Invoke: adminq_th
        '''
        
        result=True
        iterate_mp=iterate_th=True
        while iterate_mp or iterate_th: 
            module_logger.debug('Trying to read result queue')   
            
            try:
                act_result=self.__adminq_mp.get_nowait() 
            except queue.Empty:
                act_result=None
                iterate_mp=False
                result_mp=True
            else:    
                module_logger.debug("Going to play MP result: %s" % (repr(act_result)))
                result_mp=self.__play_result(act_result)
            
            try:
                act_result=self.__adminq_th.get_nowait() 
            except queue.Empty:
                act_result=None
                iterate_th=False
                result_th=True
            else:    
                module_logger.debug("Going to play TH result: %s" % (repr(act_result)))
                result_th=self.__play_result(act_result)
            
            result=result_mp and result_th
            module_logger.debug('connected result: %s, result_th: %s, result_mp: %s' % (result, result_th, result_mp))
                         
        return result
    
    def __triggers_at_task_change(self, task):
        step=self.__steps[task.step_id]
        status=task_to_step_status(task.status)
        triggers=step.triggers.get(status, None)
        triggered=list()
        if triggers:
            for event in triggers: 
                result=event.trigger_if_not_exists(self.db, task.sequence, self.__recovery)
                if result: 
                    triggered.append( (event.id_, task.sequence) )
                    module_logger.debug("Triggered post task: %s[%s]" % (repr(event.id_), task.sequence))
        return triggered
    
    def __get_admin_queue(self, task_construct):
        #if isinstance(task_construct, mp.Process):
        if task_construct == mp.Process:
            result=self.__adminq_mp
        else:
            result=self.__adminq_th
        return result
    
    def __initiate_delay(self, task, previous_task=None):
        ''' Runs delay function associated with task to register delay in delay table
        '''
        delay=self.__delays[task.step_id]
        module_logger.debug("Initiating delay: %s (previous=%s)" % (delay.delay_id, repr(previous_task)))
        active=True
        activated=datetime.utcnow()
        
        if previous_task is not None:
            prev_delay=self.__previous_delays[task.sequence][task.step_id]
            active=prev_delay.active
            activated=prev_delay.activated
            module_logger.debug("Fetched delay from previous: active: %s, activated: %s" % (active, activated))
        
        delay_func=delay.func
        try:
            result=delay_func(activated=activated, active=active, sequence=task.sequence, recovery=task.recovery)
        except Exception as e:
            task.status=TaskStatus.failure
            module_logger.critical('Exception in task execution: \n    {}'.format(task,)) #)
            trace=inspect.trace()
            trace=traces(trace)
            module_logger.critical("%s\n    %s" % (repr(e), '\n    '.join(trace)))
            module_logger.info("Stopping running processes") 
            self.__state=EventorState.shutdown
        
        if result:
            task.status=TaskStatus.success
            
        result=TaskAdminMsg(msg_type=TaskAdminMsgType.result, value=task) 
        adminq=self.__adminq_th
        adminq.put( result )

               
    def __initiate_task(self, task, previous_task=None):
        ''' Playing synchronous action.  
            
        Algorithm:
            1. Set action state to active.
            2. Launch a thread to perform action (use thread pool).
            
        Args:
            task: (Task) task to run
            previous_task: (Task), will be populated if in recovery and an instance of task exists.
        '''
        step=self.__steps[task.step_id]
        module_logger.debug("Initiating task: %s(%s)" % (task.id_, task.step_id))
        step_recovery=StepReplay.rerun
        if previous_task:
            step_recovery=step.recovery[previous_task.status]
        
        if step_recovery == StepReplay.rerun:
            # on rerun, act as before - just run the task  
            
            max_concurrent=step.config['max_concurrent']
            task_construct=step.config['task_construct']
            adminq=self.__get_admin_queue(task_construct=task_construct)
            # TODO: add join when synchronous
            #use_process=isinstance(task_construct, mp.Process)
            use_process=task_construct == mp.Process
            kwds={'run_id':self.run_id,
                  'task':task, 
                  'step': self.__steps[task.step_id], 
                  'adminq': adminq, 
                  'use_process':use_process, 
                  'logger_info':self.__logger.logger_info()}
            if max_concurrent <0 or step.concurrent < max_concurrent: # no-limit
                self.__update_task_status(task, TaskStatus.active)
                triggered=self.__triggers_at_task_change(task)
                
                #delay_task=not task.step_id.startswith('_evr_delay_')
                module_logger.debug('[ Task {} ] Going to construct ({}/{}) and run task:\n    {}'.format(task.step_id, task.sequence, task_construct, repr(task), ))
                proc=task_construct(target=task_wrapper, kwargs=kwds)
                if use_process:
                    run_id="%s-" % self.run_id if self.run_id else ''
                    proc.name = '%sTask-%s(%s)' % (run_id, step.name, task.sequence)
                step.concurrent+=1
                try:
                    proc.start()
                except Exception as e:
                    step.concurrent-=1
                    task.status=TaskStatus.failure
                    self.__update_task_status(task, TaskStatus.failure)
                    module_logger.critical('Exception in task execution: \n    {}'.format(task,)) #)
                    trace=inspect.trace()
                    trace=traces(trace)
                    module_logger.critical("%s\n    %s" % (repr(e), '\n    '.join(trace)))
                    module_logger.info("Stopping running processes") 
                    self.__state=EventorState.shutdown
                else:
                    self.__task_proc[task.id_]=proc
            else:
                module_logger.debug('[ Task {}/{} ] Delaying run (max_concurrent: {}, concurrent: {}) and run task:\n    {}'.format(task.step_id, task.sequence, max_concurrent, step.concurrent, repr(task), ))       
        else:
            # TODO: make sure htis works
            # on skip
            module_logger.debug('[ Step %s/%s ] Skipping task in recovery mode' % (task.step_id, task.sequence ))
            task.status=TaskStatus.success
            #task_result=TaskAdminMsg(msg_type=TaskAdminMsgType.result, value=task)
            triggered=self.__apply_task_result(task,)
            
        return triggered
    
    def __loop_awating_resource_allocation(self):
        while True:
            module_logger.debug("Trying to receive resources" )
            try:
                task_id=self.__rp_notify.get_nowait()
            except queue.Empty:
                break
            
            memtask=self.__tasks.get(task_id, None)    
            step=self.__steps[memtask.step_id]
            module_logger.debug("Received resources for task %s" % (step.name, ))

            memtask.resources=self.__requestors.get(memtask.request_id)
            self.__update_task_status(memtask, TaskStatus.fueled)
    
    def __update_task_status(self, task, status):
        #task.status=status    
        self.db.update_task_status(task=task, status=status)
        
    def __release_task_resources(self, task):
        step=self.__steps[task.step_id]
        module_logger.debug("Releasing task resources %s: %s" % (step.name, step.releases))
        if step.releases is not None:
            self.__requestors.put_requested(step.releases)

    def __allocate_resources_for_task(self, task, previous_task=None):
        ''' Request resources for task.  
            
        When satisfied, change status to .
        '''
        step=self.__steps[task.step_id]
        
        memtask=self.__tasks.get(task.id_, None)
        if memtask is None:
            module_logger.debug("Initiating new memtask for resource allocation %s: %s" % (step.name, step.acquires, ))
            memtask=Memtask(task)
            self.__tasks[task.id_]=memtask
        # need to allocate resources from all requested pools.
        # requests uses callback as a method to request all resources 
        # once all resources are collected, task is satisfied.
        
        #ResourceAllocationCallback(self.__resource_notification_queue)
        
        if not memtask.fueled :
            # not requested resources yet
            rp_callback=RpCallback(self.__rp_notify, task_id=task.id_)
            #requestors=vrp.Requestors(request=step.acquires, callback=rp_callback, audit=False)
            module_logger.debug("Going to reserve resources %s: %s" % (step.name, step.acquires, ))
            memtask.request_id=self.__requestors.reserve(request=step.acquires, callback=rp_callback)
            self.__update_task_status(memtask, TaskStatus.allocate)
            
    def __loop_task(self,):
        ''' evaluate ready tasks to initiate.
        
        loop_task will do its work only if eventor state is active.
        '''
        loop_seq=Sequence('_EventorTaskLoop')
        self.loop_id=loop_seq() 
        
        # all items are pulled so active can also be monitored for timely end
        # There is no need to check status.  
        #Loops will end if there is nothing to do.
        # if self.__state==EventorState.active:
        module_logger.debug("Going to fetch tasks: %s: recovery: %s" % (self.name, self.__recovery, ))
        tasks=self.db.get_task_iter(recovery=self.__recovery,status=[TaskStatus.ready, TaskStatus.allocate, TaskStatus.fueled, TaskStatus.active ])
        #module_logger.debug("Number tasks fetched: %s" % (len(list(tasks)), ))
        for task in tasks:
            step=self.__steps[task.step_id]
            try:
                previous_task=self.__previous_tasks[task.sequence][task.step_id]
            except (KeyError, TypeError):
                previous_task=None
            
            module_logger.debug("Evaluating task: %s, step: %s" % (repr(task), repr(step)))
            if task.status == TaskStatus.fueled or (task.status == TaskStatus.ready and not step.acquires):
                delay_task=task.step_id.startswith('_evr_delay_')
                if not delay_task:
                    module_logger.debug("No delay, initiate task: %s" % (task.id_, ))
                    self.__initiate_task(task, previous_task)
                else:
                    module_logger.debug("Delayed task, initiate delay for task: %s" % (task.id_, ))
                    self.__initiate_delay(task, previous_task) 
            elif task.status == TaskStatus.ready : # task.status == TaskStatus.ready and has resources to satisfy
                module_logger.debug("Ready task need to be fueled: %s" % (task.id_, ))
                self.__allocate_resources_for_task(task, previous_task)
            elif TaskStatus.allocate:
                # TODO: check if expired - if so, halt
                module_logger.debug("Task allocated resource: %s" % (task.id_, ))
            else: # active
                # TODO: check run time allowance passed - if so, halt
                pass
                    
        result=self.__collect_results()
        
        return result
            
    def __process_delay(self, db_delay):
        ''' checks Delay table 
        '''
        try:
            delay=self.__delays[db_delay.delay_id]
        except:
            module_logger.error("Delay %s not found in Delays{%s}" % (db_delay.delay_id, repr(list(self.__delays.keys()))))
            raise
        event=delay.event
        module_logger.debug("Triggering delayed event: %s(%s)" % (db_delay.delay_id, repr(event)))
        self.trigger_event(event=event, sequence=db_delay.sequence,)
        
        
    def __loop_delay(self,):
        ''' evaluate delays; act when matured.
        
        loop_delay will scan active delays.  When mature, raise its associated event.
        '''
        loop_seq=Sequence('_EventorDelayLoop')
        self.loop_id=loop_seq() 
        
        count=0
        # all items are pulled so active can also be monitored for timely end
        #if self.__state==EventorState.active:
        module_logger.debug("Going to fetch delays: %s: recovery: %s" % (self.name, self.__recovery, ))
        delays=self.db.get_delay_iter(recovery=self.__recovery,)
        now=datetime.utcnow()
        
        for delay in delays: #self.__delays.items():
            if not delay.active: continue
            age=(now-delay.activated).total_seconds()
            module_logger.debug("Delay age: %s = %s" % (delay.delay_id, age))
            if age >= delay.seconds:
                # delay is done, raise proper event.
                self.__process_delay(delay)
                self.db.deactivate_delay(delay)
            else:
                count += 1 
        
        module_logger.debug("Count of active delays: %s" % (count,))    
        return count
        
    def loop_once(self, ):
        ''' run single iteration over triggers to see if other events and associations needs to be launched.
        
        loop event: to see if new triggers matured
        loop task: to see if there is anything to run 
        '''
        
        self.__loop_delay()
        self.__loop_event()
        result = self.__loop_task()
        self.__loop_trigger_request()
        self.__loop_awating_resource_allocation()
        
        #result=todo_tasks+active_delays
        
        #module_logger.debug("Count of outstanding items: %s (to-do: %s, delays: %s)" % (result, repr(todo_tasks), repr(active_delays)))
        module_logger.debug("Loop once result: %s " % (result,))
        return result
    
    def __check_control(self):
        loop=True
        if not self.__controlq.empty():
            msg=self.__controlq.get()        
            if msg:
                loop=msg not in [LoopControl.stop]
                #self.act=msg in [LoopControl.start]
        return loop
    
    def count_todos(self, sequence=None, with_delayeds=True):
        stop_on_exception=self.__config['stop_on_exception']
        module_logger.debug('[ Step %s ] Counting todos; state: %s, recovery: %s' % (self._name(sequence), self.__state, self.__recovery))
        todo_triggers=0
        task_to_count=[TaskStatus.active, TaskStatus.fueled, TaskStatus.allocate, TaskStatus.ready,]
        
        # count triggers
        if self.__state == EventorState.active or not stop_on_exception:
            todo_triggers=self.db.count_trigger_ready(sequence=sequence, recovery=self.__recovery)
        else:
            task_to_count=[TaskStatus.active,]  
            
        # count tasks                   
        active_and_todo_tasks=self.db.count_tasks(status=task_to_count, sequence=sequence, recovery=self.__recovery) 
        total_todo=todo_triggers + active_and_todo_tasks  
        result=total_todo
        # count delay:
        if with_delayeds:
            active_delays, min_delay=self.db.count_active_delays(sequence=sequence, recovery=self.__recovery) 
            total_todo+=active_delays
            result=(total_todo, min_delay)
                             
        module_logger.debug('[ Step %s ] total todo: %s (triggers: %s, tasks: %s)' % (self._name(sequence), total_todo, todo_triggers, active_and_todo_tasks))
        
        
        return result

    def count_todos_like(self, sequence):
        ''' Counts items that are about to be in process, or in process.
        
        Args:
            sequence: (str) the sequence prefix to search for
            
        Returns:
            Count of in process 
        '''
        stop_on_exception=self.__config['stop_on_exception']
        
        module_logger.debug('[ Step %s ] Counting todos; state: %s, recovery: %s' % (self._name(sequence), self.__state, self.__recovery))
        todo_triggers=0
        task_to_count=[TaskStatus.active, TaskStatus.fueled, TaskStatus.allocate, TaskStatus.ready,]
        # If step (mega-step) is active, needs to add triggers ready.
        # else, we need to take only those tasks tat are active
        if self.__state == EventorState.active or not stop_on_exception:
            todo_triggers=self.db.count_trigger_ready_like(sequence=sequence, recovery=self.__recovery)
        else:
            # since step is not active, we are interested only with active steps with in this mega-step.
            task_to_count=[TaskStatus.active,] 
                            
        active_and_todo_tasks=self.db.count_tasks_like(status=task_to_count, sequence=sequence, recovery=self.__recovery) 
        active_delays, min_delay=self.db.count_active_delays(sequence=sequence, recovery=self.__recovery) 
        total_todo=todo_triggers + active_and_todo_tasks+active_delays                       
        module_logger.debug('[ Step %s ] total todo: %s (triggers: %s, tasks: %s)' % (self._name(sequence), total_todo, todo_triggers, active_and_todo_tasks))
        
        return (total_todo, min_delay)
    
    def loop_cycle(self,):
        ''' loops until there is no triggers or tasks.  If there are active delays, 
        they will be left for next cycle.
        
        Each loop is consist from:
            1. one processing log, 
            2. check if there is still work to do, 
            3. sleep give CPU a break
            4. check if forced to stop
        
        '''
        module_logger.debug('Starting loop cycle')
        sleep_loop=self.__config['sleep_between_loops']
        loop=True
        self.db.set_thread_synchronization(True)
        
        while loop:
            result=self.loop_once()
            # count ready triggers only if state is active
            # count ready tasks only if active
            total_todo=self.count_todos(with_delayeds=False)                       
            loop=total_todo>0 
            if loop:
                loop=self.__check_control()
                if loop:
                    time.sleep(sleep_loop)
                    loop=self.__check_control()
                if not loop and not self._session_cycle_loop:
                    module_logger.info('Processing stopped')
            else:
                pass
            
        return result

    def loop_session(self):
        ''' loops until there is no work to do
        
        Each loop is consist from:
            1. one processing log, 
            2. check if there is still work to do, 
            3. sleep give CPU a break
            4. check if forced to stop
        
        '''
        module_logger.debug('Starting loop session')
        sleep_loop=self.__config['sleep_between_loops']
        loop=True
        self.db.set_thread_synchronization(True)
        self._session_cycle_loop=True
        while loop:
            result=self.loop_cycle()
            #result=self.loop_once()
            # count ready triggers only if state is active
            # count ready tasks only if active
            total_todo, min_delay=self.count_todos()                       
            loop=total_todo>0 
            if loop:
                loop=self.__check_control()
                if loop:
                    sleep_time=sleep_loop if min_delay is None else min_delay
                    if sleep_loop != sleep_time:
                        module_logger.debug('Making a time delay sleep: %s' % sleep_time)
                    time.sleep(sleep_time) # time.sleep(sleep_loop)
                    loop=self.__check_control()
                if not loop:
                    module_logger.info('Processing stopped, number outstanding tasks: %s' % total_todo)
            else:
                human_result="success" if result else 'failure'
                module_logger.info('Processing finished with: %s' % human_result)
                
        #self.__logger.stop()  
        result=self.__state != EventorState.shutdown   
        return result
    
    def loop_session_stop(self):
        self.__controlq.put(LoopControl.stop)
        
    def get_step_sequence(self):
        result=os.environ.get('EVENTOR_STEP_SEQUENCE', '')
        return result
    
    def get_step_name(self):
        result=os.environ.get('EVENTOR_STEP_NAME', '')
        return result
    
    def get_task_status(self, task_names, sequence,):
        result=self.db.get_task_status(task_names=task_names, sequence=sequence, recovery=self.__recovery)
        return result
        
    def run(self,  max_loops=-1):
        ''' loops events structures to execute raise events and execute tasks.
        
        Args:
            max_loops: number of loops to run.  If positive, limits number of loops.
                 defaults to negative, which would run loops until there are no events to raise and
                 no task to run. 
        
        '''
        
        if setproctitle is not None:
            run_id = "%s." % self.run_id if self.run_id else ''
            setproctitle("eventor: %s" % (run_id,))
        
        if max_loops < 0:
            result=self.loop_session()
        else:
            result=None
            for _ in range(max_loops):
                #module_logger.debug('Starting loop cycle')
                result=self.loop_cycle()
            human_result="success" if result else 'failure'
            total_todo, _=self.count_todos(with_delayeds=True) 
            module_logger.info('Processing finished with: %s; outstanding tasks: %s' % (human_result, total_todo))
            #module_logger.info('Processing finished')
                
        return result
    
    
    def close(self):
        ''' closes open artifacts like MPlogger etc.
        '''
        if self.__logger is not None: 
            #print("stopping logger")
            try:
                self.__logger.stop()
            except AttributeError:
                pass



'''
Created on Jan 31, 2014

@author: arnon
'''
import os

_varprog = None
_varprogb = None

def expandvars(source, environ=None):
    """Expand shell variables of form $var and ${var}.  Unknown variables
    are left unchanged."""
    if source is None:
        return source
    if environ is None:
        environ=os.environ
        #os.path.expandvars(path)
    global _varprog, _varprogb
    if isinstance(source, bytes):
        if b'$' not in source:
            return source
        if not _varprogb:
            import re
            #_varprogb = re.compile(br'\$(\w+|\{[^}]*\})', re.ASCII)
            _varprogb = re.compile(br'\$(\w+|\{[^}]*\})')
        search = _varprogb.search
        start = b'{'
        end = b'}'
    elif isinstance(source, str):
        if '$' not in source:
            return source
        if not _varprog:
            import re
            #_varprog = re.compile(r'\$(\w+|\{[^}]*\})', re.ASCII)
            _varprog = re.compile(r'\$(\w+|\{[^}]*\})')
        search = _varprog.search
        start = '{'
        end = '}'
    else:
        return source
    i = 0
    while True:
        m = search(source, i)
        if not m:
            break
        i, j = m.span(0)
        name = m.group(1)
        if name.startswith(start) and name.endswith(end):
            name = name[1:-1]
        if isinstance(name, bytes):
            name = str(name)
        if name in environ.keys():
            tail = source[j:]
            value = str(environ[name])
            try:
                parts=value.split('\\')
            except Exception as e:
                print(name, value, ':', repr(e))
                raise
            value=os.path.join(*parts)
            if isinstance(source, bytes):
                value = value.encode('ASCII')
            source = source[:i] + value
            i = len(source)
            source += tail
        else:
            i = j
    #return os.path.normpath(source)    
    return source

def pathhasvars(source):
    """Expand shell variables of form $var and ${var}.  Unknown variables
    are left unchanged."""
    global _varprog, _varprogb
    if isinstance(source, bytes):
        if b'$' not in source:
            return False
        if not _varprogb:
            import re
            #_varprogb = re.compile(br'\$(\w+|\{[^}]*\})', re.ASCII)
            _varprogb = re.compile(br'\$(\w+|\{[^}]*\})')
        search = _varprogb.search
        start = b'{'
        end = b'}'
    else:
        if '$' not in source:
            return False
        if not _varprog:
            import re
            #_varprog = re.compile(r'\$(\w+|\{[^}]*\})', re.ASCII)
            _varprog = re.compile(r'\$(\w+|\{[^}]*\})')
        search = _varprog.search
        start = '{'
        end = '}'
    i = 0
    m = search(source, i)
    return not (not m)  
